# -*- coding: utf-8 -*-
# Modified work Copyright (c) 2017-2018 Science and Technology
# Facilities Council.
# Original work Copyright (c) 1999-2008 Pearu Peterson.

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# --------------------------------------------------------------------

# The original software (in the f2py project) was distributed under
# the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the F2PY project nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

"""
Fortran block statements.

"""

import logging
import re
import sys

import fparser.one.statements as statements
from fparser.one.statements import *
import fparser.one.typedecl_statements as typedecl_statements
from fparser.one.typedecl_statements import *

from fparser.common.base_classes import BeginStatement, EndStatement, \
                                        Statement, AttributeHolder,   \
                                        ProgramBlock, Variable
from fparser.common.utils import split_comma, filter_stmts, parse_bind, \
                                 parse_result, AnalyzeError, is_name

__all__ = ['BeginSource', 'Module', 'PythonModule', 'Program', 'BlockData',
           'Interface', 'Subroutine', 'Function', 'SelectCase', 'SelectType',
           'WhereConstruct', 'ForallConstruct', 'IfThen', 'If', 'Do',
           'Associate', 'TypeDecl', 'Enum', 'EndSource', 'EndModule',
           'EndPythonModule', 'EndProgram', 'EndBlockData', 'EndInterface',
           'EndSubroutine', 'EndFunction', 'EndSelect', 'EndWhere',
           'EndForall', 'EndIfThen', 'EndDo', 'EndAssociate', 'EndType',
           'EndEnum']
__all__.extend(statements.__all__)
__all__.extend(typedecl_statements.__all__)


class HasImplicitStmt(object):
    '''
    Class encapsulating information about any Implicit statements
    contained within a scoping block.
    '''
    a = AttributeHolder(implicit_rules={})

    def get_type_by_name(self, name):
        '''
        Returns an object of the correct type (Integer or Real) using
        Fortran's implicit typing rules for the supplied variable name.

        :param str name: The variable name.
        :returns: Object describing the variable.
        :rtype: Either :py:class:`fparser.one.typedecl_statements.Real` \
                or :py:class:`fparser.one.typedecl_statements.Integer`.

        '''
        # The implicit_rules dict is populated by the analyze() method
        # of one.typedecl_statements.Implicit
        implicit_rules = self.a.implicit_rules
        if implicit_rules is None:
            raise AnalyzeError('Implicit rules mapping is null '
                               'while getting %r type' % (name))
        line = name[0].lower()
        if line in implicit_rules:
            return implicit_rules[line]
        # default rules:
        if line in 'ijklmn':
            line = 'default_integer'
        else:
            line = 'default_real'
        var = implicit_rules.get(line, None)
        if var is None:
            if line[8:] == 'real':
                implicit_rules[line] = var = Real(self, self.item.copy('real'))
            else:
                implicit_rules[line] = var = Integer(self,
                                                     self.item.copy('integer'))
        return var

    def topyf(self, tab='  '):
        '''
        Constructs a pyf representation of this class.

        :param str tab: White space to prepend to output.
        :returns: pyf code for this implicit statement.
        :rtype: str
        '''
        implicit_rules = self.a.implicit_rules
        if implicit_rules is None:
            return tab + 'IMPLICIT NONE\n'
        # Construct a dict where the keys are types and the items are
        # the list of initial letters mapped to that type
        items = {}
        for char, itype in list(implicit_rules.items()):
            if char.startswith('default'):
                continue
            type_str = itype.tostr()
            if type_str in items:
                items[type_str].append(char)
            else:
                items[type_str] = [char]
        if not items:
            return tab + '! default IMPLICIT rules apply\n'
        stmt = 'IMPLICIT'
        impl_list = []
        for itype, letter_list in list(items.items()):
            letter_list.sort()
            impl_list.append(itype + ' (%s)' % (', '.join(letter_list)))
        stmt += ' ' + ', '.join(impl_list)
        return tab + stmt + '\n'


class HasUseStmt(object):

    a = AttributeHolder(use={},
                        use_provides={})

    def get_entity(self, name):
        for modname, modblock in list(self.top.a.module.items()):
            for stmt in modblock.content:
                if getattr(stmt, 'name', '') == name:
                    return stmt
        return

    def topyf(self, tab='  '):
        sys.stderr.write('HasUseStmt.topyf not implemented\n')
        return ''


class AccessSpecs(object):

    a = AttributeHolder(private_id_list=[], public_id_list=[])

    def topyf(self, tab='  '):
        private_list = self.a.private_id_list
        public_list = self.a.public_id_list
        lines = []
        if '' in private_list:
            lines.append(tab + 'PRIVATE\n')
        if '' in public_list:
            lines.append(tab + 'PUBLIC\n')
        for a in private_list:
            if not a:
                continue
            lines.append(tab + 'PRIVATE :: %s\n' % (a))
        for a in public_list:
            if not a:
                continue
            lines.append(tab + 'PUBLIC :: %s\n' % (a))
        return ''.join(lines)


class HasVariables(object):

    a = AttributeHolder(variables={},
                        variable_names=[]  # defines the order of declarations
                        )

    def get_variable_by_name(self, name):
        variables = self.a.variables
        if name in variables:
            var = variables[name]
        else:
            var = variables[name] = Variable(self, name)
            self.a.variable_names.append(name)
        return var

    def topyf(self, tab='', only_variables=None):
        s = ''
        if only_variables is None:
            only_variables = list(self.a.variables.keys())
        for name in only_variables:
            var = self.a.variables[name]
            s += tab + str(var) + '\n'
        return s


class HasTypeDecls(object):

    a = AttributeHolder(type_decls={})

    def topyf(self, tab=''):
        s = ''
        for name, stmt in list(self.a.type_decls.items()):
            s += stmt.topyf(tab='  '+tab)
        return s

    def get_type_decl_by_kind(self, kind):
        type_decls = self.a.type_decls
        type_decl = type_decls.get(kind, None)
        if type_decl is None:
            return self.get_entity(kind)
        return type_decl


class HasAttributes(object):

    known_attributes = []
    a = AttributeHolder(attributes=[])

    def topyf(self, tab=''):
        s = ''
        for attr in self.a.attributes:
            s += tab + attr + '\n'
        return s

    def update_attributes(self, *attrs):
        attributes = self.a.attributes
        known_attributes = self.known_attributes
        if len(attrs) == 1 and isinstance(attrs[0], (tuple, list)):
            attrs = attrs[0]
        for attr in attrs:
            uattr = attr.upper()
            if uattr not in attributes:
                if isinstance(known_attributes, (list, tuple)):
                    if uattr not in known_attributes:
                        self.warning('unknown attribute %r' % (attr))
                elif not known_attributes(uattr):
                    self.warning('unknown attribute %r' % (attr))
                attributes.append(uattr)
            else:
                self.warning('multiple specification of attribute %r' % (attr))
        return


class HasModuleProcedures(object):

    a = AttributeHolder(module_procedures=[])


# File block

class EndSource(EndStatement):
    """
    Dummy End statement for BeginSource.
    """
    match = staticmethod(lambda s: False)


class BeginSource(BeginStatement):
    """
    Fortran source content.
    """
    match = staticmethod(lambda s: True)
    end_stmt_cls = EndSource
    a = AttributeHolder(module={},
                        external_subprogram={},
                        blockdata={},
                        )

    def tofortran(self, isfix=None):
        if isfix:
            tab = 'C'
        else:
            tab = self.get_indent_tab(isfix=isfix) + '!'
        return tab + BeginStatement.tofortran(self, isfix=isfix)

    def tostr(self):
        return self.blocktype.upper() + ' ' + self.name

    def process_item(self):
        self.name = self.reader.name
        self.top = self
        self.fill(end_flag=True)
        return

    def analyze(self):
        for stmt in self.content:
            if isinstance(stmt, Module):
                stmt.analyze()
                self.a.module[stmt.name] = stmt
            elif isinstance(stmt, SubProgramStatement):
                stmt.analyze()
                self.a.external_subprogram[stmt.name] = stmt
            elif isinstance(stmt, BlockData):
                stmt.analyze()
                self.a.blockdata[stmt.name] = stmt
            else:
                stmt.analyze()
        return

    def get_classes(self):
        if self.reader.format.is_pyf:
            return [PythonModule] + program_unit
        return program_unit

    def process_subitem(self, item):
        # MAIN block does not define start/end line conditions,
        # so it should never end until all lines are read.
        # However, sometimes F77 programs lack the PROGRAM statement,
        # and here we fix that:
        if self.reader.format.is_f77:
            line = item.get_line()
            if line == 'end':
                message = item.reader.format_message(
                    'WARNING',
                    'assuming the end of undefined PROGRAM statement',
                    item.span[0], item.span[1])
                logger.warning(message)
                # print >> sys.stderr, message
                p = Program(self)
                p.content.extend(self.content)
                p.content.append(EndProgram(p, item))
                self.content[:] = [p]
                return
        return BeginStatement.process_subitem(self, item)

    def topyf(self, tab=''):  # XXXX
        s = ''
        for name, stmt in list(self.a.module.items()):
            s += stmt.topyf(tab=tab)
        for name, stmt in list(self.a.external_subprogram.items()):
            s += stmt.topyf(tab=tab)
        for name, stmt in list(self.a.blockdata.items()):
            s += stmt.topyf(tab=tab)
        return s


# Module

class EndModule(EndStatement):
    match = re.compile(r'end(\s*module\s*\w*|)\Z', re.I).match


class Module(BeginStatement, HasAttributes,
             HasImplicitStmt, HasUseStmt, HasVariables,
             HasTypeDecls, AccessSpecs):
    """
    MODULE <name>
     ..
    END [MODULE [name]]
    """
    match = re.compile(r'module\s*\w+\Z', re.I).match
    end_stmt_cls = EndModule

    a = AttributeHolder(module_subprogram={},
                        module_provides={},  # all symbols that are public and
                                             # so can be imported via USE
                                             # statement by other blocks
                        module_interface={}
                        )

    known_attributes = ['PUBLIC', 'PRIVATE']

    def get_classes(self):
        return access_spec + specification_part + module_subprogram_part

    def process_item(self):
        name = self.item.get_line().replace(' ', '')
        name = name[len(self.blocktype):].strip()
        self.name = name
        return BeginStatement.process_item(self)

    def get_provides(self):
        return self.a.module_provides

    def get_interface(self):
        return self.a.module_interface

    def analyze(self):
        content = self.content[:]

        while content:
            stmt = content.pop(0)
            if isinstance(stmt, Contains):
                for stmt in filter_stmts(content, SubProgramStatement):
                    stmt.analyze()
                    self.a.module_subprogram[stmt.name] = stmt
                stmt = content.pop(0)
                while isinstance(stmt, Comment):
                    stmt = content.pop(0)
                if not isinstance(stmt, EndModule):
                    stmt.error('Expected END MODULE statement (analyzer).')
                continue
            stmt.analyze()

        if content:
            logger.info('Not analyzed content: %s' % content)
            # self.show_message('Not analyzed content: %s' % content)

        module_provides = self.a.module_provides
        for name, var in list(self.a.variables.items()):
            if var.is_public():
                if name in module_provides:
                    message = 'module data object name conflict with %s, ' \
                              + 'overriding.'
                    self.warning(message % (name))
                module_provides[name] = var

        return

    def topyf(self, tab=''):
        s = tab + 'MODULE '+self.name + '\n'
        s += HasImplicitStmt.topyf(self, tab=tab+'  ')
        s += AccessSpecs.topyf(self, tab=tab+'  ')
        s += HasAttributes.topyf(self, tab=tab+'  ')
        s += HasTypeDecls.topyf(self, tab=tab+'  ')
        s += HasVariables.topyf(self, tab=tab+'  ')
        for name, stmt in list(self.a.module_interface.items()):
            s += stmt.topyf(tab=tab+'    ')
        s += tab + '  CONTAINS\n'
        for name, stmt in list(self.a.module_subprogram.items()):
            s += stmt.topyf(tab=tab+'    ')
        s += tab + 'END MODULE ' + self.name + '\n'
        return s

    def check_private(self, name):
        if name in self.a.public_id_list:
            return False
        if name in self.a.private_id_list:
            return True
        if '' in self.a.public_id_list:
            return False
        if '' in self.a.private_id_list:
            return True
        # todo: handle generic-spec-s in id-lists.
        return


# Python Module

class EndPythonModule(EndStatement):
    match = re.compile(r'end(\s*python\s*module\s*\w*|)\Z', re.I).match


class PythonModule(BeginStatement, HasImplicitStmt, HasUseStmt,
                   HasVariables):
    """
    PYTHON MODULE <name>
     ..
    END [PYTHON MODULE [name]]
    """
    modes = ['pyf']
    match = re.compile(r'python\s*module\s*\w+\Z', re.I).match
    end_stmt_cls = EndPythonModule

    def get_classes(self):
        return [Interface, Function, Subroutine, Module]

    def process_item(self):
        self.name = self.item.get_line().replace(' ', '')
        self.name = self.name[len(self.blocktype):].strip()
        return BeginStatement.process_item(self)


# Program

class EndProgram(EndStatement):
    """
    END [PROGRAM [name]]
    """
    match = re.compile(r'end(\s*program\s*\w*|)\Z', re.I).match


class Program(BeginStatement, ProgramBlock,
              # HasAttributes, # XXX: why Program needs .attributes?
              HasVariables,
              HasImplicitStmt, HasUseStmt, AccessSpecs):
    """ PROGRAM [name]
    """
    match = re.compile(r'program\s*\w*\Z', re.I).match
    end_stmt_cls = EndProgram

    def get_classes(self):
        return specification_part + execution_part + internal_subprogram_part

    def process_item(self):
        if self.item is not None:
            name = self.item.get_line().replace(' ', '')
            name = name[len(self.blocktype):].strip()
            if name:
                self.name = name
        return BeginStatement.process_item(self)


# BlockData

class EndBlockData(EndStatement):
    """
    END [BLOCK DATA [<block-data-name>]]
    """
    match = re.compile(r'end(\s*block\s*data\s*\w*|)\Z', re.I).match
    blocktype = 'blockdata'


class BlockData(BeginStatement, HasImplicitStmt, HasUseStmt,
                HasVariables, AccessSpecs):
    """
    BLOCK DATA [<block-data-name>]
    """
    end_stmt_cls = EndBlockData
    match = re.compile(r'block\s*data\s*\w*\Z', re.I).match

    def process_item(self):
        self.name = self.item.get_line()[5:].lstrip()[4:].lstrip()
        return BeginStatement.process_item(self)

    def get_classes(self):
        return specification_part


# Interface

class EndInterface(EndStatement):
    match = re.compile(r'end\s*interface\s*(\w+\s*\(.*\)|\w*)\Z', re.I).match
    blocktype = 'interface'


class Interface(BeginStatement, HasAttributes, HasImplicitStmt, HasUseStmt,
                HasModuleProcedures, AccessSpecs
                ):
    """
    INTERFACE [<generic-spec>] | ABSTRACT INTERFACE
    END INTERFACE [<generic-spec>]

    <generic-spec> = <generic-name>
                   | OPERATOR ( <defined-operator> )
                   | ASSIGNMENT ( = )
                   | <dtio-generic-spec>
    <dtio-generic-spec> = READ ( FORMATTED )
                        | READ ( UNFORMATTED )
                        | WRITE ( FORMATTED )
                        | WRITE ( UNFORMATTED )

    """
    modes = ['free', 'fix', 'pyf']
    pattern = r'(interface\s*(\w+\s*\(.*\)|\w*)|abstract\s*interface)\Z'
    match = re.compile(pattern, re.I).match
    end_stmt_cls = EndInterface
    blocktype = 'interface'

    a = AttributeHolder(interface_provides={})

    def get_classes(self):
        line = intrinsic_type_spec + interface_specification
        if self.reader.format.mode == 'pyf':
            return [Subroutine, Function] + line
        return line

    def process_item(self):
        line = self.item.get_line()
        line = self.item.apply_map(line)
        self.isabstract = line.startswith('abstract')
        if self.isabstract:
            self.generic_spec = ''
        else:
            self.generic_spec = line[len(self.blocktype):].strip()
        self.name = self.generic_spec  # XXX
        return BeginStatement.process_item(self)

    def tostr(self):
        if self.isabstract:
            return 'ABSTRACT INTERFACE'
        return 'INTERFACE ' + str(self.generic_spec)

    def analyze(self):
        content = self.content[:]

        while content:
            stmt = content.pop(0)
            if isinstance(stmt, self.end_stmt_cls):
                break
            stmt.analyze()
        if content:
            logger.info('Not analyzed content: %s' % content)
            # self.show_message('Not analyzed content: %s' % content)

        if self.name in self.parent.a.variables:
            var = self.parent.a.variables.pop(self.name)
            self.update_attributes(var.attributes)

        if isinstance(self.parent, Module):  # XXX
            parent_interface = self.parent.get_interface()
            if self.name in parent_interface:
                p = parent_interface[self.name]
                last = p.content.pop()
                assert isinstance(last, EndInterface), repr(last.__class__)
                p.content += self.content
                p.update_attributes(self.a.attributes)
            else:
                parent_interface[self.name] = self
            return

    def topyf(self, tab=''):
        s = tab + self.tostr() + '\n'
        s += HasImplicitStmt.topyf(self, tab=tab+'  ')
        s += HasAttributes.topyf(self, tab=tab+'  ')
        s += HasUseStmt.topyf(self, tab=tab+'  ')
        s += tab + 'END' + self.tostr() + '\n'
        return s


# Subroutine

class SubProgramStatement(BeginStatement, ProgramBlock,
                          HasImplicitStmt, HasAttributes,
                          HasUseStmt,
                          HasVariables, HasTypeDecls, AccessSpecs
                          ):
    """
    [<prefix>] <FUNCTION|SUBROUTINE> <name> [( <args> )] [<suffix>]
    """

    a = AttributeHolder(internal_subprogram={})
    known_attributes = ['RECURSIVE', 'PURE', 'ELEMENTAL']

    def process_item(self):
        clsname = self.__class__.__name__.lower()
        item = self.item
        line = item.get_line()
        m = self.match(line)
        i = line.lower().find(clsname)
        assert i != -1, repr((clsname, line))
        self.prefix = line[:i].rstrip()
        self.name = line[i:m.end()].lstrip()[len(clsname):].strip()
        line = line[m.end():].lstrip()
        args = []
        if line.startswith('('):
            i = line.find(')')
            assert i != -1, repr(line)
            line2 = item.apply_map(line[:i+1])
            for a in line2[1:-1].split(','):
                a = a.strip()
                if not a:
                    continue
                args.append(a)
            line = line[i+1:].lstrip()
        suffix = item.apply_map(line)
        self.bind, suffix = parse_bind(suffix, item)
        self.result = None
        if isinstance(self, Function):
            self.result, suffix = parse_result(suffix, item)
            if suffix:
                assert self.bind is None, repr(self.bind)
                self.bind, suffix = parse_result(suffix, item)
            if self.result is None:
                self.result = self.name
        assert not suffix, repr(suffix)
        self.args = args
        self.typedecl = None
        return BeginStatement.process_item(self)

    def tostr(self):
        clsname = self.__class__.__name__.upper()
        s = ''
        if self.prefix:
            s += self.prefix + ' '
        if self.typedecl is not None:
            assert isinstance(self, Function), repr(self.__class__.__name__)
            s += self.typedecl.tostr() + ' '
        s += clsname
        suf = ''
        if self.result and self.result != self.name:
            suf += ' RESULT ( %s )' % (self.result)
        if self.bind:
            suf += ' BIND ( %s )' % (', '.join(self.bind))
        return '%s %s(%s)%s' % (s, self.name, ', '.join(self.args), suf)

    def get_classes(self):
        return f2py_stmt + specification_part + execution_part \
               + internal_subprogram_part

    def analyze(self):
        content = self.content[:]

        if self.prefix:
            self.update_attributes(self.prefix.upper().split())

        variables = self.a.variables
        for a in self.args:
            assert a not in variables
            if is_name(a):
                variables[a] = Variable(self, a)
            elif a == '*':
                variables[a] = Variable(self, a)  # XXX: fix me appropriately
            else:
                message = 'argument must be a name or * but got %r'
                raise AnalyzeError(message % (a))

        if isinstance(self, Function):
            var = variables[self.result] = Variable(self, self.result)
            if self.typedecl is not None:
                var.set_type(self.typedecl)

        while content:
            stmt = content.pop(0)
            if isinstance(stmt, Contains):
                for stmt in filter_stmts(content, SubProgramStatement):
                    stmt.analyze()
                    self.a.internal_subprogram[stmt.name] = stmt
                stmt = content.pop(0)
                while isinstance(stmt, Comment):
                    stmt = content.pop(0)
                assert isinstance(stmt, self.end_stmt_cls), repr(stmt)
            elif isinstance(stmt, self.end_stmt_cls):
                continue
            else:
                if hasattr(stmt, "analyze"):
                    stmt.analyze()
                else:
                    message = 'Failed to parse: {0}'
                    raise AnalyzeError(message.format(str(stmt)))

        if content:
            logger.info('Not analyzed content: %s' % content)
            # self.show_message('Not analyzed content: %s' % content)

        parent_provides = self.parent.get_provides()
        if parent_provides is not None:
            if self.name in parent_provides:
                message = 'module subprogram name conflict with %s, ' \
                          + 'overriding.'
                self.warning(message % (self.name))
            if self.is_public():
                parent_provides[self.name] = self

        if self.is_recursive() and self.is_elemental():
            message = 'C1241 violation: prefix cannot specify both ' \
                      + 'ELEMENTAL and RECURSIVE'
            self.warning(message)
        return

    def topyf(self, tab=''):
        s = tab + self.__class__.__name__.upper()
        s += ' ' + self.name + ' (%s)' % (', '.join(self.args))
        if isinstance(self, Function) and self.result != self.name:
            s += ' RESULT (%s)' % (self.result)
        s += '\n'
        s += HasImplicitStmt.topyf(self, tab=tab+'  ')
        s += AccessSpecs.topyf(self, tab=tab+'  ')
        s += HasTypeDecls.topyf(self, tab=tab+'  ')
        s += HasVariables.topyf(self, tab=tab + '  ',
                                only_variables=self.args)
        s += tab + 'END ' + self.__class__.__name__.upper() \
            + ' ' + self.name + '\n'
        return s

    def is_public(self): return not self.is_private()

    def is_private(self): return self.parent.check_private(self.name)

    def is_recursive(self): return 'RECURSIVE' in self.a.attributes

    def is_pure(self): return 'PURE' in self.a.attributes

    def is_elemental(self): return 'ELEMENTAL' in self.a.attributes


class EndSubroutine(EndStatement):
    """
    END [SUBROUTINE [name]]
    """
    match = re.compile(r'end\s*(?:subroutine\s*(?:(?P<name>\w+)\s*)?)?$',
                       re.IGNORECASE).match


class Subroutine(SubProgramStatement):
    """
    [<prefix>] SUBROUTINE <name> [( [<dummy-arg-list>] )
    [<proc-language-binding-spec>]]
    """
    end_stmt_cls = EndSubroutine
    pattern = r'(recursive|pure|elemental|\s)*subroutine\s*\w+'
    match = re.compile(pattern, re.I).match
    _repr_attr_names = ['prefix', 'bind', 'suffix', 'args'] \
        + Statement._repr_attr_names


# Function

class EndFunction(EndStatement):
    """
    END [FUNCTION [name]]
    """
    match = re.compile(r'end(\s*function\s*\w*|)\Z', re.I).match


class Function(SubProgramStatement):
    """
    [<prefix>] FUNCTION <name> ( [<dummy-arg-list>] ) [<suffix>]
    <prefix> = <prefix-spec> [<prefix-spec>]...
    <prefix-spec> = <declaration-type-spec>
                  | RECURSIVE | PURE | ELEMENTAL
    <suffix> = <proc-language-binding-spec> [RESULT ( <result-name> )]
             | RESULT ( <result-name> ) [<proc-language-binding-spec>]
    """
    end_stmt_cls = EndFunction
    pattern = r'(recursive|pure|elemental|\s)*function\s*\w+'
    match = re.compile(pattern, re.I).match
    _repr_attr_names = ['prefix', 'bind', 'suffix', 'args', 'typedecl'] \
        + Statement._repr_attr_names

    def subroutine_wrapper_code(self):
        name = 'f2pywrap_' + self.name
        args = ['f2pyvalue_'+self.result] + self.args
        var = self.a.variables[self.result]
        typedecl = var.get_typedecl().astypedecl()
        lines = []
        tab = ' '*6
        lines.append('%sSUBROUTINE %s(%s)' % (tab, name, ', '.join(args)))
        if isinstance(self.parent, Module):
            lines.append('%s  USE %s' % (tab, self.parent.name))
        else:
            if isinstance(typedecl, TypeStmt):
                type_decl = typedecl.get_type_decl(typedecl.name)
                if type_decl.parent is self:
                    for line in str(type_decl).split('\n'):
                        lines.append('%s  %s' % (tab, line.lstrip()))
            lines.append('%s  EXTERNAL %s' % (tab, self.name))
            lines.append('%s  %s %s' % (tab,
                                        str(typedecl).lstrip(),
                                        self.name))
        lines.append('%s  %s %s' % (tab, str(typedecl).lstrip(), args[0]))
        lines.append('!f2py intent(out) %s' % (args[0]))
        for a in self.args:
            v = self.a.variables[a]
            lines.append('%s  %s' % (tab, str(v).lstrip()))
        lines.append('%s  %s = %s(%s)' % (tab,
                                          args[0],
                                          self.name,
                                          ', '.join(self.args)))
        lines.append('%sEND SUBROUTINE %s' % (tab, name))
        return '\n'.join(lines)

    def subroutine_wrapper(self):
        code = self.subroutine_wrapper_code()
        from .api import parse
        block = parse(code)  # XXX: set include_dirs
        while len(block.content) == 1:
            block = block.content[0]
        return block


# Handle subprogram prefixes

class SubprogramPrefix(Statement):
    """
    <prefix> <declaration-type-spec> <function|subroutine> ...
    """
    match = re.compile(r'(pure|elemental|recursive|\s)+\b', re.I).match

    def process_item(self):
        line = self.item.get_line()
        m = self.match(line)
        prefix = line[:m.end()].rstrip()
        rest = self.item.get_line()[m.end():].lstrip()
        if rest:
            self.parent.put_item(self.item.copy(prefix))
            self.item.clone(rest)
            self.isvalid = False
            return
        if self.parent.__class__ not in [Function, Subroutine]:
            self.isvalid = False
            return
        prefix = prefix + ' ' + self.parent.prefix
        self.parent.prefix = prefix.strip()
        self.ignore = True
        return

# SelectCase


class EndSelect(EndStatement):
    match = re.compile(r'end\s*select\s*\w*\Z', re.I).match
    blocktype = 'select'


class Select(BeginStatement):
    """
    Base class for the Select (case/type) statement

    """
    end_stmt_cls = EndSelect
    name = ''

    def process_item(self):
        ''' Populate the state of this Select object by parsing the
        associated line of code '''
        item = self.item
        # TODO make the following more robust, particularly to the
        # presence of a name at the beginning
        # (e.g. "a_name: select case(...)")
        line = item.get_line()[6:].lstrip()[4:].lstrip()[1:-1].strip()
        self.expr = item.apply_map(line)
        self.construct_name = self.item.name
        return BeginStatement.process_item(self)


class SelectCase(Select):
    """
    [<case-construct-name> :] SELECT CASE ( <case-expr> )

    """
    match = re.compile(r'select\s*case\s*\(.*\)\Z', re.I).match

    def tostr(self):
        return 'SELECT CASE ( %s )' % (self.expr)

    def get_classes(self):
        ''' Return the list of classes that this instance may
        have as children '''
        return [Case] + execution_part_construct


class SelectType(Select):
    """
    [<case-construct-name> :] SELECT TYPE ( <case-expr> )

    """
    match = re.compile(r'select\s*type\s*\(.*\)\Z', re.I).match

    def tostr(self):
        return 'SELECT TYPE ( %s )' % (self.expr)

    def get_classes(self):
        ''' Return the list of classes that this instance may
        have as children '''
        return [TypeIs, ClassIs] + execution_part_construct


# Where

class EndWhere(EndStatement):
    """
    END WHERE [<where-construct-name>]
    """
    match = re.compile(r'end\s*\where\s*\w*\Z', re.I).match


class Where(BeginStatement):
    """
    [<where-construct-name> :] WHERE ( <mask-expr> )
    <mask-expr> = <logical-expr>
    """
    match = re.compile(r'where\s*\([^)]*\)\Z', re.I).match
    end_stmt_cls = EndWhere
    name = ''

    def tostr(self):
        return 'WHERE ( %s )' % (self.expr)

    def process_item(self):
        self.expr = self.item.get_line()[5:].lstrip()[1:-1].strip()
        self.construct_name = self.item.name
        return BeginStatement.process_item(self)

    def get_classes(self):
        return [Assignment, WhereStmt, WhereConstruct, ElseWhere]


WhereConstruct = Where


# Forall

class EndForall(EndStatement):
    """
    END FORALL [<forall-construct-name>]
    """
    match = re.compile(r'end\s*forall\s*\w*\Z', re.I).match


class Forall(BeginStatement):
    """
    [<forall-construct-name> :] FORALL <forall-header>
      [<forall-body-construct>]...
    <forall-body-construct> = <forall-assignment-stmt>
                            | <where-stmt>
                            | <where-construct>
                            | <forall-construct>
                            | <forall-stmt>
    <forall-header> = ( <forall-triplet-spec-list> [, <scalar-mask-expr>] )
    <forall-triplet-spec> = <index-name> = <subscript> : <subscript>
                            [: <stride>]
    <subscript|stride> = <scalar-int-expr>
    <forall-assignment-stmt> = <assignment-stmt> | <pointer-assignment-stmt>
    """
    end_stmt_cls = EndForall
    match = re.compile(r'forall\s*\(.*\)\Z', re.I).match
    name = ''

    def process_item(self):
        self.specs = self.item.get_line()[6:].lstrip()[1:-1].strip()
        return BeginStatement.process_item(self)

    def tostr(self):
        return 'FORALL (%s)' % (self.specs)

    def get_classes(self):
        return [GeneralAssignment, WhereStmt, WhereConstruct,
                ForallConstruct, ForallStmt]


ForallConstruct = Forall


# IfThen

class EndIfThen(EndStatement):
    """
    END IF [<if-construct-name>]
    """
    match = re.compile(r'end\s*if\s*\w*\Z', re.I).match
    blocktype = 'if'


class IfThen(BeginStatement):
    """
    [<if-construct-name> :] IF ( <scalar-logical-expr> ) THEN

    IfThen instance has the following attributes:
      expr
    """

    match = re.compile(r'if\s*\(.*\)\s*then\Z', re.I).match
    end_stmt_cls = EndIfThen
    name = ''

    def tostr(self):
        return 'IF (%s) THEN' % (self.expr)

    def process_item(self):
        item = self.item
        line = item.get_line()[2:-4].strip()
        assert line[0] == '(' and line[-1] == ')', repr(line)
        self.expr = item.apply_map(line[1:-1].strip())
        self.construct_name = item.name
        return BeginStatement.process_item(self)

    def get_classes(self):
        return [Else, ElseIf] + execution_part_construct


class If(BeginStatement):
    """
    IF ( <scalar-logical-expr> ) action-stmt
    """

    match = re.compile(r'if\s*\(', re.I).match

    def process_item(self):
        item = self.item
        mode = self.reader.format.mode
        classes = self.get_classes()
        classes = [cls for cls in classes if mode in cls.modes]

        line = item.get_line()[2:].lstrip()
        i = line.find(')')
        expr = line[1:i].strip()
        line = line[i+1:].strip()
        if line.lower() == 'then':
            self.isvalid = False
            return
        self.expr = item.apply_map(expr)

        if not line:
            newitem = self.get_item()
        else:
            newitem = item.copy(line, apply_map=True)
        newline = newitem.get_line()
        for cls in classes:
            if cls.match(newline):
                stmt = cls(self, newitem)
                if stmt.isvalid:
                    self.content.append(stmt)
                    return
        if not line:
            self.put_item(newitem)
        self.isvalid = False
        return

    def tostr(self):
        assert len(self.content) == 1, repr(self.content)
        return 'IF (%s) %s' % (self.expr, str(self.content[0]).lstrip())

    def tofortran(self, isfix=None):
        return self.get_indent_tab(isfix=isfix) + self.tostr()

    def get_classes(self):
        return action_stmt


# Do

class EndDo(EndStatement):
    """
    END DO [<do-construct-name>]
    """
    match = re.compile(r'end\s*do\s*(?:(?P<name>\w+)\s*)?\Z',
                       re.IGNORECASE).match
    blocktype = 'do'

    def process_item(self):
        '''
        Parses the next line assuming it is an "End do" statement.

        Overrides method in `EndStatement`.
        '''
        item = self.item
        line = item.get_line()
        matched = self.match(line)
        # Check for matching labels
        found_label = getattr(self.item, 'label', None)
        expected_label = getattr(self.parent, 'endlabel', None)
        if expected_label:
            if found_label:
                if found_label != expected_label:
                    message = 'When entering the "do" block {start} was' \
                              + ' given as the end label but {end} was found.'
                    self.warning(message.format(start=expected_label,
                                                end=found_label))
                    self.isvalid = False
            else:
                message = 'A label was specified when entering the "do" ' \
                          + ' block ({label}) but none was found at the end.'
                self.warning(message.format(label=expected_label))
                self.isvalid = False
        # Check for matching names
        found_name = matched.group('name') or None
        expected_name = self.parent.construct_name
        if expected_name:
            if found_name:
                if found_name != expected_name:
                    message = 'The "do" block was specified with the name' \
                              + ' "{open}" but was closed with the name' \
                              + ' "{close}".'
                    self.warning(message.format(open=expected_name,
                                                close=found_name))
                    self.isvalid = False
            else:
                message = 'A name ("{name}") was specified for the "do" ' \
                          + 'block but was not given when closing the block.'
                self.warning(message.format(name=expected_name))
                self.isvalid = False
        else:
            if found_name:
                message = 'The name "{name}" was used when closing a "do"' \
                          + 'block but none was specified when opening it.'
                self.warning(message.format(name=found_name))
                self.isvalid = False
        return EndStatement.process_item(self)


class Do(BeginStatement):
    """
    [<do-construct-name> :] DO label [loopcontrol]
    [<do-construct-name> :] DO [loopcontrol]

    """

    match = re.compile(r'do\b\s*\d*', re.I).match
    pattern = r'do\b\s*(?:(?P<label>\d+)\s*)?(?:,\s*)?(?P<loopcontrol>.+)?\Z'
    item_re = re.compile(pattern, re.IGNORECASE).match
    end_stmt_cls = EndDo
    name = ''

    def tostr(self):
        lst = ['DO']
        for part in [self.endlabel, self.loopcontrol]:
            if part:
                lst.append(str(part))
        return ' '.join(lst)

    def process_item(self):
        '''
        Parses the next line assuming it is a "Do" statement.

        Overrides method in `BeginStatement`.
        '''
        item = self.item
        line = item.get_line()
        matched = self.item_re(line)
        if matched.group('label'):
            self.endlabel = int(matched.group('label').strip())
        else:
            self.endlabel = None
        self.construct_name = item.name
        if matched.group('loopcontrol'):
            self.loopcontrol \
                = item.apply_map(matched.group('loopcontrol').strip())
        else:
            self.loopcontrol = None
        return BeginStatement.process_item(self)

    def process_subitem(self, item):
        result = False
        if self.endlabel:
            label = item.label
            if label == self.endlabel:
                result = True
                if isinstance(self.parent,
                              Do) and label == self.parent.endlabel:
                    # the same item label may be used for different block ends
                    self.put_item(item)
        return BeginStatement.process_subitem(self, item) or result

    def get_classes(self):
        return execution_part_construct


# Associate

class EndAssociate(EndStatement):
    """
    END ASSOCIATE [<associate-construct-name>]
    """
    match = re.compile(r'end\s*associate\s*\w*\Z', re.I).match


class Associate(BeginStatement):
    """
    [<associate-construct-name> :] ASSOCIATE ( <association-list> )
      <block>

    <association> = <associate-name> => <selector>
    <selector> = <expr> | <variable>
    """
    match = re.compile(r'associate\s*\(.*\)\Z', re.I).match
    end_stmt_cls = EndAssociate

    def process_item(self):
        line = self.item.get_line()[9:].lstrip()
        self.associations = line[1:-1].strip()
        return BeginStatement.process_item(self)

    def tostr(self):
        return 'ASSOCIATE (%s)' % (self.associations)

    def get_classes(self):
        return execution_part_construct


# Type

class EndType(EndStatement):
    """
    END TYPE [<type-name>]
    """
    match = re.compile(r'end\s*type\s*\w*\Z', re.I).match
    blocktype = 'type'


class Type(BeginStatement, HasVariables, HasAttributes, HasModuleProcedures,
           AccessSpecs):
    """
    TYPE [[, <typ-attr-spec-list>] ::] <type-name> [( <type-param-name-list> )]
    <typ-attr-spec> = <access-spec> | EXTENDS ( <parent-type-name> )
                      | ABSTRACT | BIND(C)
    """
    match = re.compile(r'type\b\s*').match
    end_stmt_cls = EndType

    a = AttributeHolder(extends=None,
                        parameters={},
                        component_names=[],  # Order for sequence types
                        components={}
                        )
    pattern = r'\A(PUBLIC|PRIVATE|SEQUENCE|ABSTRACT|BIND\s*\(.*\))\Z'
    known_attributes = re.compile(pattern, re.I).match

    def process_item(self):
        line = self.item.get_line()[4:].lstrip()
        if line.startswith('('):
            self.isvalid = False
            return
        specs = []
        i = line.find('::')
        if i != -1:
            for s in line[:i].split(','):
                s = s.strip()
                if s:
                    specs.append(s)
            line = line[i+2:].lstrip()
        self.specs = specs
        i = line.find('(')
        if i != -1:
            self.name = line[:i].rstrip()
            assert line[-1] == ')', repr(line)
            self.params = split_comma(line[i+1:-1].lstrip())
        else:
            self.name = line
            self.params = []
        if not is_name(self.name):
            self.isvalid = False
            return
        return BeginStatement.process_item(self)

    def tostr(self):
        s = 'TYPE'
        if self.specs:
            s += ', '.join(['']+self.specs) + ' ::'
        s += ' ' + self.name
        if self.params:
            s += ' ('+', '.join(self.params)+')'
        return s

    def get_classes(self):
        return [Integer] + private_or_sequence + component_part +\
               type_bound_procedure_part

    def analyze(self):
        for spec in self.specs:
            i = spec.find('(')
            if i != -1:
                assert spec.endswith(')'), repr(spec)
                s = spec[:i].rstrip().upper()
                n = spec[i+1:-1].strip()
                if s == 'EXTENDS':
                    self.a.extends = n
                    continue
                elif s == 'BIND':
                    args, rest = parse_bind(spec)
                    assert not rest, repr(rest)
                    spec = 'BIND(%s)' % (', '.join(args))
                else:
                    spec = '%s(%s)' % (s, n)
            else:
                spec = spec.upper()
            self.update_attributes(spec)

        component_names = self.a.component_names
        content = self.content[:]
        while content:
            stmt = content.pop(0)
            if isinstance(stmt, self.end_stmt_cls):
                break
            stmt.analyze()

        if content:
            message = 'Not analyzed content: %s' % content
            logging.getLogger(__class__).info(message)

        parameters = self.a.parameters
        components = self.a.components
        component_names = self.a.component_names
        for name in self.a.variable_names:
            var = self.a.variables[name]
            if name in self.params:
                parameters[name] = var
            else:
                component_names.append(name)
                components[name] = var

        self.parent.a.type_decls[self.name] = self

        parent_provides = self.parent.get_provides()
        if parent_provides is not None:
            if self.is_public():
                if self.name in parent_provides:
                    message = 'type declaration name conflict with {}, ' \
                              + 'overriding.'
                    self.warning(message.format(self.name))
                parent_provides[self.name] = self

        return

    def topyf(self, tab=''):
        s = tab + 'TYPE'
        if self.a.extends is not None:
            s += ', EXTENDS(%s) ::' % (self.a.extends)
        s += ' ' + self.name
        if self.a.parameters:
            s += ' (%s)' % (', '.join(self.a.parameters))
        s += '\n'
        s += AccessSpecs.topyf(self, tab=tab+'  ')
        s += HasAttributes.topyf(self, tab=tab+'  ')
        s += HasVariables.topyf(self, tab=tab+'  ')
        s += tab + 'END TYPE ' + self.name + '\n'
        return s

    # Wrapper methods:

    def is_public(self): return not self.is_private()

    def is_private(self):
        if 'PUBLIC' in self.a.attributes:
            return False
        if 'PRIVATE' in self.a.attributes:
            return True
        return self.parent.check_private(self.name)


TypeDecl = Type


# Enum

class EndEnum(EndStatement):
    """
    END ENUM
    """
    match = re.compile(r'end\s*enum\Z', re.I).match
    blocktype = 'enum'


class Enum(BeginStatement):
    """
    ENUM , BIND(C)
      <enumerator-def-stmt>
      [<enumerator-def-stmt>]...
    """
    blocktype = 'enum'
    end_stmt_cls = EndEnum
    match = re.compile(r'enum\s*,\s*bind\s*\(\s*c\s*\)\Z', re.I).match

    def process_item(self):
        return BeginStatement.process_item(self)

    def get_classes(self):
        return [Enumerator]


##############################################################################

f2py_stmt = [Threadsafe, FortranName, Depend, Check, CallStatement,
             CallProtoArgument]

access_spec = [Public, Private]

interface_specification = [Function, Subroutine, ModuleProcedure]

module_subprogram_part = [Contains, Function, Subroutine]

specification_stmt = access_spec \
    + [Allocatable, Asynchronous, Bind, Common, Data, Dimension, Equivalence,
       External, Intent, Intrinsic, Namelist, Optional, Pointer, Protected,
       Save, Target, Volatile, Value]

intrinsic_type_spec = [SubprogramPrefix, Integer, Real,  DoublePrecision,
                       Complex, DoubleComplex, Character, Logical, Byte]

derived_type_spec = []
type_spec = intrinsic_type_spec + derived_type_spec
declaration_type_spec = intrinsic_type_spec + [TypeStmt, Class]

type_declaration_stmt = declaration_type_spec

private_or_sequence = [Private, Sequence]

component_part = declaration_type_spec + [ModuleProcedure]

proc_binding_stmt = [SpecificBinding, GenericBinding, FinalBinding]

type_bound_procedure_part = [Contains, Private] + proc_binding_stmt

# R214
action_stmt = [Allocate, GeneralAssignment, Assign, Backspace, Call, Close,
               Continue, Cycle, Deallocate, Endfile, Exit, Flush, ForallStmt,
               Goto, If, Inquire, Nullify, Open, Print, Read, Return, Rewind,
               Stop, Wait, WhereStmt, Write, ArithmeticIf, ComputedGoto,
               AssignedGoto, Pause]
# GeneralAssignment = Assignment + PointerAssignment
# EndFunction, EndProgram, EndSubroutine - part of the corresponding blocks

executable_construct = [Associate, Do, ForallConstruct, IfThen,
                        SelectCase, SelectType, WhereConstruct] + action_stmt

execution_part_construct = executable_construct + [Format, Entry,
                                                   Data]

execution_part = execution_part_construct[:]

# C201, R208
for cls in [EndFunction, EndProgram, EndSubroutine]:
    try:
        execution_part.remove(cls)
    except ValueError:
        pass

internal_subprogram = [Function, Subroutine]

internal_subprogram_part = [Contains] + internal_subprogram

# In Fortran2003 we can have a Procedure declaration. We therefore
# include SpecificBinding as a valid declaration construct.
declaration_construct = [TypeDecl, Entry, Enum, Format, Interface,
                         Parameter, ModuleProcedure, SpecificBinding] + \
    specification_stmt + type_declaration_stmt
# stmt-function-stmt

implicit_part = [Implicit, Parameter, Format, Entry]

specification_part = [Use, Import] + implicit_part + \
                     declaration_construct


external_subprogram = [Function, Subroutine]

main_program = [Program] + specification_part + execution_part + \
               internal_subprogram_part

program_unit = main_program + external_subprogram + [Module,
                                                     BlockData]



# Copyright (c) 2019 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test Fortran 2003 : This file tests the support for a Hollerith
string. Hollerith strings take the form `nHx`, where `n` is an integer
and `x` is a sequence of characters of length `n`.

Note, the Hollerith format was deprecated in Fortran77 and removed in
Fortran95. However, Fortran compilers still support it.

'''

import pytest
from fparser.two.Fortran2003 import Hollerith_Item
from fparser.two.utils import NoMatchError, InternalError


def test_hollerith(f2003_create, monkeypatch):
    '''Check that a valid hollerith string is parsed correctly.'''
    from fparser.two import utils
    monkeypatch.setattr(utils, "EXTENSIONS", ["hollerith"])
    for myinput in ["2Hab", "  2Hab", "1h ", "7h1234567", " 1 1 H01234567890"]:
        ast = Hollerith_Item(myinput)
        expected = myinput.upper()
        # Remove any spaces before the H (but not after).
        lhs, rhs = expected.split('H')
        expected = lhs.replace(' ', '')+'H'+rhs
        assert str(ast).upper() == expected


def test_repr(f2003_create, monkeypatch):
    '''Check that the repr output of a hollerith string gives the expected
    result.

    '''
    from fparser.two import utils
    monkeypatch.setattr(utils, "EXTENSIONS", ["hollerith"])
    myinput = "2Hab"
    ast = Hollerith_Item(myinput)
    assert repr(ast) == "Hollerith_Item('ab')"


def test_syntaxerror(f2003_create, monkeypatch):
    '''test that an exception is raised if the supplied format is
    invalid.

    '''
    from fparser.two import utils
    monkeypatch.setattr(utils, "EXTENSIONS", ["hollerith"])
    for myinput in [None, "", "  ", "0H", "1H", "2Hx" "2Hxxx", "H20", "xH"]:
        with pytest.raises(NoMatchError):
            _ = Hollerith_Item(myinput)


def test_internal_error1(f2003_create, monkeypatch):
    '''Check that an internal error is raised if the length of the Items
    list is not 1 as the str() method assumes that it is.

    '''
    from fparser.two import utils
    monkeypatch.setattr(utils, "EXTENSIONS", ["hollerith"])
    myinput = "2Hab"
    ast = Hollerith_Item(myinput)
    monkeypatch.setattr(ast, "items", [None, None])
    with pytest.raises(InternalError) as excinfo:
        str(ast)
    assert "should be of length 1 but found '2'" in str(excinfo.value)


def test_internal_error2(f2003_create, monkeypatch):
    '''Check that an internal error is raised if the string value (entry 0
    of Items) is empty or None as the str() method assumes that it is
    a string with content.

    '''
    from fparser.two import utils
    monkeypatch.setattr(utils, "EXTENSIONS", ["hollerith"])
    myinput = "2hab"
    ast = Hollerith_Item(myinput)
    for content in [None, ""]:
        monkeypatch.setattr(ast, "items", [content])
        with pytest.raises(InternalError) as excinfo:
            str(ast)
        assert ("entry 0 should be a valid Hollerith string but it is "
                "empty") in str(excinfo.value)


def test_invalid_hollerith(f2003_create, monkeypatch):
    '''Test that the hollerith extension to the standard raises an
    exception if it is not named as a valid extension.

    '''
    from fparser.two import utils
    monkeypatch.setattr(utils, "EXTENSIONS", [])
    myinput = "2Hab"
    with pytest.raises(NoMatchError) as excinfo:
        _ = Hollerith_Item(myinput)
    assert "Hollerith_Item: '{0}'".format(myinput) \
        in str(excinfo.value)



# Copyright (c) 2017-2020 Science and Technology Facilities Council.

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

''' Module containing pytest tests for fparser2 base classes '''

import pytest


@pytest.mark.usefixtures("f2003_create")
def test_keywordvaluebase_errors():
    ''' Unit tests for the KeywordValueBase class to check that it rejects
    invalid input '''
    from fparser.two.Fortran2003 import KeywordValueBase, Io_Unit, Format, \
        Char_Literal_Constant
    lhs_cls = 'UNIT'
    rhs_cls = Io_Unit
    obj = KeywordValueBase.match(lhs_cls, rhs_cls, "  ", require_lhs=False)
    assert obj is None
    obj = KeywordValueBase.match(lhs_cls, rhs_cls, " = 36 ",)
    assert obj is None
    _, obj = KeywordValueBase.match("FMT", Format, "'(A)'", require_lhs=False)
    assert isinstance(obj, Char_Literal_Constant)
    _, obj = KeywordValueBase.match("FMT", Format, "FMT='(A)'",
                                    require_lhs=False)
    assert isinstance(obj, Char_Literal_Constant)
    _, obj = KeywordValueBase.match("FMT", Format,
                                    "FMT='(\"my_var =  \", (A))'",
                                    require_lhs=False)
    assert isinstance(obj, Char_Literal_Constant)
    lhs, obj = KeywordValueBase.match("FMT", Format,
                                      "FMT='(\"my_var =  \", (A))'",
                                      require_lhs=True)
    assert lhs == "FMT"
    assert isinstance(obj, Char_Literal_Constant)
    # Try to trigger an error by specifying that a "LHS =" is required but
    # then omitting it while having the RHS contain a '=' character
    obj = KeywordValueBase.match("FMT", Format,
                                 "'(\"my_var =  \", (A))'",
                                 require_lhs=True)
    assert obj is None


@pytest.mark.usefixtures("f2003_create")
def test_read_stmt_errors():
    ''' Unit tests for the Read class to ensure it rejects invalid
    inputs '''
    from fparser.two.Fortran2003 import Read_Stmt
    # Missing closing parenthesis
    obj = Read_Stmt.match("READ(unit=23")
    assert obj is None
    # Missing arguments
    obj = Read_Stmt.match("READ()")
    assert obj is None
    obj = Read_Stmt.match("READ")
    assert obj is None
    # Wrong argument type
    obj = Read_Stmt.match("READ a_var")
    assert obj is None
    obj = Read_Stmt.match("READ 13")
    assert obj is None
    # Missing comma
    obj = Read_Stmt.match("READ * a_var")
    assert obj is None
    # Missing value/variable after comma
    obj = Read_Stmt.match("READ 13, ")
    assert obj is None


@pytest.mark.usefixtures("f2003_create")
def test_io_ctrl_spec_list_errors():
    ''' Unit tests for the Io_Control_Spec_List class to ensure it
    rejects invalid input '''
    from fparser.two.Fortran2003 import Io_Control_Spec_List
    # Positional arg following named arg
    obj = Io_Control_Spec_List.match("unit=23, namvar")
    assert obj is None


@pytest.mark.usefixtures("f2003_create")
def test_io_ctrl_spec_errors():
    ''' Unit tests for the Io_Control_Spec class to ensure it
    rejects invalid input '''
    from fparser.two.Fortran2003 import Io_Control_Spec
    # An argument with a name that is not valid within an IO control
    # description
    obj = Io_Control_Spec.match("not_unit=23")
    assert obj is None


@pytest.mark.usefixtures("f2003_create")
def test_blockbase_tofortran_non_ascii():
    ''' Check that the tofortran() method works when we have a program
    containing non-ascii characters within a sub-class of BlockBase. We
    use a Case Construct for this purpose. '''
    from fparser.common.readfortran import FortranStringReader
    from fparser.two.utils import BlockBase, walk
    from fparser.two.Fortran2003 import Program, Case_Construct
    code = (u"program my_test\n"
            u"! A comment outside the select block\n"
            u"SELECT CASE(iflag)\n"
            u"CASE(  30  )\n"
            u"  IF(lwp) WRITE(*,*) ' for e1=1\xb0'\n"
            u"END SELECT\n"
            u"end program\n")
    reader = FortranStringReader(code, ignore_comments=False)
    obj = Program(reader)
    bbase = walk(obj.content, Case_Construct)[0]
    # Explicitly call tofortran() on the BlockBase class.
    out_str = BlockBase.tofortran(bbase)
    assert "for e1=1" in out_str



# This Python file uses the following encoding: utf-8
# Copyright (c) 2020 Science and Technology Facilities Council.

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

''' pytest module for the Fortran2003 Where Construct - R744. Note that
    this requires re-implementing and extending to fully cover the
    language specification (#232). '''

import pytest
from fparser.api import get_reader
from fparser.common.readfortran import FortranStringReader
from fparser.two.Fortran2003 import Where_Construct


@pytest.mark.usefixtures("f2003_create")
def test_where_construct():
    ''' Tests for the WHERE construct, R744. '''
    tcls = Where_Construct
    obj = tcls(get_reader('''\
    where (pressure <= 1.0)
    pressure = pressure + inc_pressure
    temp = temp - 5.0
    elsewhere
    raining = .true.
    end where
'''))
    assert isinstance(obj, tcls), repr(obj)
    assert (str(obj) == "WHERE (pressure <= 1.0)\n  "
            "pressure = pressure + inc_pressure\n  "
            "temp = temp - 5.0\n"
            "ELSEWHERE\n  raining = .TRUE.\nEND WHERE")

    obj = tcls(get_reader('''\
    where (cond1)
    else    where (cond2)
    end where
'''))
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == 'WHERE (cond1)\nELSEWHERE(cond2)\nEND WHERE'

    obj = tcls(get_reader('''\
    n:where (cond1)
    elsewhere (cond2) n
    else   where n
    end where n
'''))
    assert isinstance(obj, tcls), repr(obj)
    assert (str(obj) == "n:WHERE (cond1)\nELSEWHERE(cond2) n\n"
            "ELSEWHERE n\nEND WHERE n")

    obj = tcls(get_reader('''\
    n:where (cond1)
    else where (cond2) n
    else where n
    end where n
'''))
    assert isinstance(obj, tcls), repr(obj)
    assert (str(obj) ==
            'n:WHERE (cond1)\nELSEWHERE(cond2) n\nELSEWHERE n\n'
            'END WHERE n')

    obj = tcls(get_reader('''\
    n:where (me(:)=="hello")
    else where (me(:)=="goodbye") n
    else where n
    end where n
'''))
    assert (str(obj) ==
            'n:WHERE (me(:) == "hello")\nELSEWHERE(me(:) == "goodbye") n\n'
            'ELSEWHERE n\n'
            'END WHERE n')


@pytest.mark.usefixtures("f2003_create")
def test_where_tofortran_non_ascii():
    ''' Check that the tofortran() method works when a WHERE construct
    contains a character string with non-ascii characters. '''
    code = (u"WHERE(iflag)\n"
            u"  msg_list = ' for e1=1\xb0'\n"
            u"END WHERE\n")
    reader = FortranStringReader(code, ignore_comments=False)
    obj = Where_Construct(reader)
    out_str = str(obj)
    assert "for e1=1" in out_str



# Copyright (c) 2019 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test Fortran 2003 rules R703 and R723 : This file tests the support
for a Defined Unary Operator and a Defined Binary Operator,
respectively. Both rules are included in a single test as, in terms of
fparser implementation, they are identical. Therefore, to avoid code
duplication we use a fixture.

'''

import pytest
from fparser.two.utils import NoMatchError


def test_defined_operator(f2003_create, op_type):
    '''Check that basic unary and binary operators are parsed correctly.

    '''
    for line in [".myoperator.",
                 "  .myoperator.  ",
                 "."+63*"a"+"."]:
        ast = op_type(line)
        target = line.strip().upper()
        assert target in str(ast)
        assert repr(ast) == "Defined_Op('{0}')".format(target)


def test_syntax_error(f2003_create, op_type):
    ''' Test that NoMatchError is raised for various syntax errors. '''
    for line in ["", "  ", "x", ".x", "x.", "..", ".,.", ". x.", ".x .",
                 ".x x."]:
        with pytest.raises(NoMatchError) as excinfo:
            _ = op_type(line)
        assert "_Op: '{0}'".format(line) in str(excinfo.value)


def test_c703(f2003_create, op_type):
    '''Test that we do not match if the name is more than 63 characters
    long, the name matches an existing intrinsic operator (e.g. .gt.),
    or the name matches an existing logical literal constant
    (e.g. .false.).

    '''
    for line in ["."+64*"a"+".", ".eq.", ".not.", ".false.", ".FALSE."]:
        with pytest.raises(NoMatchError) as excinfo:
            _ = op_type(line)
        assert "_Op: '{0}'".format(line) in str(excinfo.value)



# Copyright (c) 2019 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test Fortran 2003 rule R759 : This file tests the support for the
forall statement.

'''

import pytest
from fparser.api import get_reader
from fparser.two.Fortran2003 import Forall_Stmt
from fparser.two.utils import NoMatchError, InternalError


def test_forall_stmt(f2003_create):
    '''Check that a basic forall statement is parsed correctly. Input
    separately as a string and as a reader object.

    '''
    def check_forall(reader):
        '''Internal helper function to avoid code replication.'''
        ast = Forall_Stmt(reader)
        assert "FORALL (i = 1 : 2, j = 1 : 2) a(i, j) = i + j" in str(ast)

    for line in ["forall (i=1:2,j=1:2) a(i,j)=i+j",
                 "FoRaLl (i=1:2,j=1:2) a(i,j)=i+j",
                 "  forall  (  i=1:2,j=1:2  )   a(i,j)=i+j  "]:
        check_forall(line)
        reader = get_reader(line)
        check_forall(reader)


def test_forall_stmt_brackets(f2003_create):
    '''Check that a basic forall statement with brackets in the triplet
    list is parsed correctly. Input separately as a string and as a
    reader object.

    '''
    def check_forall(reader):
        '''Internal helper function to avoid code replication.'''
        ast = Forall_Stmt(reader)
        assert "FORALL (i = b(k) : 2, j = 1 : c(k)) a(i, j) = 0.0" in str(ast)

        line = "forall (i=b(k):2,j=1:c(k)) a(i,j)=0.0"
        check_forall(line)
        reader = get_reader(line)
        check_forall(reader)


def test_syntaxerror(f2003_create):
    '''Test that NoMatchError is raised for various syntax errors.'''
    for line in ["", "   ", "forall",
                 "for all (i=1,2,j=1:2) a(i,j)=i+j",
                 "forl (i=1,2,j=1:2) a(i,j)=i+j",
                 "forall i=1,2,j=1:2) a(i,j)=i+j",
                 "forall (i=1,2,j=1:2 a(i,j)=i+j",
                 "forall i=1,2,j=1:2) a=0.0",
                 "forall (i=1,2,j=1:2 a=0.0",
                 "forall x (i=1,2,j=1:2) a(i,j)=i+j",
                 "forall (i=1,2,j=1:2) x a(i,j)=i+j",
                 "forall (i=1,2,j=1:2)",
                 "forall a(i,j)=i+j",
                 "forall () a(i,j)=i+j"]:
        with pytest.raises(NoMatchError) as excinfo:
            _ = Forall_Stmt(line)
        assert "Forall_Stmt: '{0}'".format(line) in str(excinfo.value)


def test_internal_error1(f2003_create, monkeypatch):
    '''Check that an internal error is raised if the length of the Items
    list is not 2 as the str() method assumes that it is.

    '''
    line = "forall (i=1:2,j=1:2) a(i,j)=0.0"
    ast = Forall_Stmt(line)
    monkeypatch.setattr(ast, "items", [None, None, None])
    with pytest.raises(InternalError) as excinfo:
        str(ast)
    assert ("Class Forall_Stmt method tostr() has '3' items, but "
            "expecting 2.") in str(excinfo.value)


def test_internal_error2(f2003_create):
    '''Check that an internal error is raised if entry 0 of Items is empty
    or None as the str() method assumes that it is a string with
    content.

    '''
    line = "forall (i=1:2,j=1:2) a(i,j)=0.0"
    ast = Forall_Stmt(line)
    for content in [None, ""]:
        ast.items = (content, ast.items[1])
        with pytest.raises(InternalError) as excinfo:
            str(ast)
        assert ("'Items' entry 0 should be a valid "
                "Forall_Header") in str(excinfo.value)


def test_internal_error3(f2003_create):
    '''Check that an internal error is raised if entry 1 of Items is empty
    or None as the str() method assumes that it is a string with
    content.

    '''
    line = "forall (i=1:2,j=1:2) a(i,j)=0.0"
    ast = Forall_Stmt(line)
    for content in [None, ""]:
        ast.items = (ast.items[0], content)
        with pytest.raises(InternalError) as excinfo:
            str(ast)
        assert ("'Items' entry 1 should be a valid "
                "Forall_Assignment_Stmt") in str(excinfo.value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Modified work Copyright (c) 2017-2020 Science and Technology
# Facilities Council
# Original work Copyright (c) 1999-2008 Pearu Peterson

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# --------------------------------------------------------------------

# The original software (in the f2py project) was distributed under
# the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the F2PY project nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.
#
# Author: Pearu Peterson <pearu@cens.ioc.ee>
# Created: May 2006
# Modified by R. W. Ford, STFC Daresbury Lab
# Modified by P. Elson, Met Office

"""Provides Fortran reader classes.

Overview
========

Provides FortranReader classes for reading Fortran codes from files and
strings. FortranReader handles comments and line continuations of both
fix and free format Fortran codes.

Examples
========

::

    >> from fparser.common.readfortran import FortranFileReader
    >>> import os
    >>> reader = FortranFileReader(os.path.expanduser('~/src/blas/daxpy.f'))
    >>> print reader.next()
    line #1 'subroutine daxpy(n,da,dx,incx,dy,incy)'
    >>> print `reader.next()`
    Comment('c     constant times a vector plus a vector.\\n
    c     uses unrolled loops for increments equal to one.\\n
    c     jack dongarra, linpack, 3/11/78.\\n
    c     modified 12/3/93, array(1) declarations changed to array(*)',(3, 6))
    >>> print `reader.next()`
    Line('double precision dx(*),dy(*),da',(8, 8),'')
    >>> print `reader.next()`
    Line('integer i,incx,incy,ix,iy,m,mp1,n',(9, 9),'')

Note that the ``.next()`` method may return `Line`, `SyntaxErrorLine`,
`Comment`, `MultiLine`, or `SyntaxErrorMultiLine` instance.
Let us continue with the above example session to illustrate the `Line`
methods and attributes::

    >>> item = reader.next()
    >>> item
        Line('if (da .eq. 0.0d0) return',(12, 12),'')
    >>> item.line
        'if (da .eq. 0.0d0) return'
    >>> item.strline
        'if (F2PY_EXPR_TUPLE_5) return'
    >>> item.strlinemap
        {'F2PY_EXPR_TUPLE_5': 'da .eq. 0.0d0'}
    >>> item.span
        (12, 12)
    >>> item.get_line()
        'if (F2PY_EXPR_TUPLE_5) return'

To read a Fortran code from a string, use `FortranStringReader` class::

    >>> from fparser.common.sourceinfo import FortranFormat
    >>> from fparser.common.readfortran import FortranStringReader
    >>> code = '''
    ...       subroutine foo(a)
    ...         integer a
    ...         print*,\"a=\",a
    ...       end
    ... '''
    >>> reader = FortranStringReader(code)
    >>> reader.set_format(FortranFormat(False, True))
    >>> reader.next()
        Line('subroutine foo(a)',(2, 2),'')
    >>> reader.next()
        Line('integer a',(3, 3),'')
    >>> reader.next()
        Line('print*,\"a=\",a',(4, 4),'')

"""

from __future__ import print_function

import logging
import io
import os
import re
import sys
import traceback
import six
import fparser.common.sourceinfo
from fparser.common.splitline import String, string_replace_map, splitquote


__all__ = ['FortranFileReader',
           'FortranStringReader',
           'FortranReaderError',
           'Line',
           'SyntaxErrorLine',
           'Comment',
           'MultiLine',
           'SyntaxErrorMultiLine']

_SPACEDIGITS = ' 0123456789'
_CF2PY_RE = re.compile(r'(?P<indent>\s*)!f2py(?P<rest>.*)', re.I)
_LABEL_RE = re.compile(r'\s*(?P<label>\d+)\s*(\b|(?=&)|\Z)', re.I)
_CONSTRUCT_NAME_RE = re.compile(r'\s*(?P<name>\w+)\s*:\s*(\b|(?=&)|\Z)', re.I)
_IS_INCLUDE_LINE = re.compile(r'\s*include\s*("[^"]+"'
                              + r'|\'[^\']+\')\s*\Z', re.I).match


def _is_fix_cont(line):
    return line and len(line) > 5 and line[5] != ' ' and line[:5] == 5 * ' '


def _is_fix_comment(line, isstrict):
    """
    Check whether line is a comment line in fixed format Fortran source.

    :param str line: Line of code to check
    :param bool isstrict: Whether we are strictly enforcing fixed/free fmt

    References
    ----------
    :f2008:`3.3.3`
    """
    if line:
        if line[0] in '*cC!':
            return True
        if not isstrict:
            i = line.find('!')
            if i != -1:
                start = line[:i].lstrip()
                if not start:
                    if i == 5:
                        # line continuation
                        return False
                    return True
                else:
                    # inline comment or ! is used in character context
                    # inline comments are handled elsewhere
                    pass
    elif line == '':
        return True
    return False


_HOLLERITH_START_SEARCH = re.compile(r'(?P<pre>\A|,\s*)'
                                     + r'(?P<num>\d+)h', re.I).search
_IS_CALL_STMT = re.compile(r'call\b', re.I).match


def extract_label(line):
    '''Look for an integer label at the start of 'line' and if there is
    one then remove it from 'line' and store it as an integer in
    'label', returning both in a tuple.

    :param str line: a string that potentially contains a label at the \
        start.

    :returns: a 2-tuple containing the label and updated line if a \
        label is found or None and the unchanged line if a label is \
        not found.
    :rtype: (int or NoneType, str)

    '''
    label = None
    match = _LABEL_RE.match(line)
    if match:
        label = int(match.group('label'))
        line = line[match.end():].lstrip()
    return label, line


def extract_construct_name(line):
    '''Look for a construct name at the start of 'line' and if there is
    one then remove it from 'line' and return it as a string in
    'name', returning both in a tuple.

    :param str line: a string that potentially contains a construct \
        name at the start.

    :returns: a 2-tuple containing the construct name and updated line \
        if a construct name is found or None and the unchanged line if \
        a construct name is not found.
    :rtype: (str or NoneType, str)

    '''
    construct_name = None
    match = _CONSTRUCT_NAME_RE.match(line)
    if match:
        construct_name = match.group('name')
        line = line[match.end():].lstrip()
    return construct_name, line


class FortranReaderError(Exception):
    '''
    Thrown when there is an error reading the Fortran source file.
    '''
    pass


class Line(object):
    """ Holds a Fortran source line.

    Attributes
    ----------
    line : str
      code line
    span : 2-tuple
      starting and ending line numbers
    label : {int, None}
      Specify statement label
    name : {str, None}
      Specify construct name.
    reader : FortranReaderBase
    strline : {None, str}
    is_f2py_directive : bool
      the line contains f2py directive
    """

    f2py_strmap_findall = re.compile(r'(_F2PY_STRING_CONSTANT_\d+_'
                                     + r'|F2PY_EXPR_TUPLE_\d+)').findall

    def __init__(self, line, linenospan, label, name, reader):
        self.line = line.strip()
        if not self.line:
            raise FortranReaderError("Got empty line: '{0}'. linenospan={1}, "
                                     "label='{2}'".format(line,
                                                          linenospan, label))
        self.span = linenospan
        assert label is None or isinstance(label, int), repr(label)
        assert name is None or isinstance(name, str) and name != '', repr(name)
        self.label = label
        self.name = name
        self.reader = reader
        self.strline = None
        self.is_f2py_directive = linenospan[0] in reader.f2py_comment_lines
        self.parse_cache = {}

    def has_map(self):
        '''
        Returns true when a substitution map has been registered.
        '''
        return hasattr(self, 'strlinemap') and self.strlinemap

    def apply_map(self, line):
        '''
        Substitutes magic strings in a line with values specified in a map.
        '''
        if not hasattr(self, 'strlinemap') or not self.strlinemap:
            return line
        findall = self.f2py_strmap_findall
        str_map = self.strlinemap
        keys = findall(line)
        for k in keys:
            line = line.replace(k, str_map[k])
        return line

    def copy(self, line=None, apply_map=False):
        '''
        Creates a Line object from a string.

        If no line argument is specified a copy is made of this Line.

        If a substitution map is provided it is used while making the copy.
        '''
        if line is None:
            line = self.line
        if apply_map:
            line = self.apply_map(line)
        return Line(line, self.span, self.label, self.name, self.reader)

    def clone(self, line):
        '''
        This Line has its contents overwitten by the passed string. The
        incoming string has substitution applied.
        '''
        self.line = self.apply_map(line)
        self.strline = None
        return

    def __repr__(self):
        return self.__class__.__name__+'(%r,%s,%r,%r,<reader>)' \
               % (self.line, self.span, self.label, self.name)

    def __str__(self):
        s = 'line #%s' % (self.span[0])
        if self.label is not None:
            s += ' %s ' % (self.label)
        if self.name is not None:
            s += '%s: ' % (self.name)
        return s + repr(self.line)

    def isempty(self, ignore_comments=False):
        return not (self.line
                    or self.label is not None
                    or self.name is not None)

    def get_line(self, apply_map=False):
        if apply_map:
            return self.apply_map(self.get_line(apply_map=False))
        if self.strline is not None:
            return self.strline
        line = self.line

        if self.reader.format.is_f77:
            # Handle Hollerith constants by replacing them
            # with char-literal-constants.
            # H constants may appear only in DATA statements and
            # in the argument list of CALL statement.
            # Hollerith constants were removed from the Fortran 77 standard.
            # The following handling is not perfect but works for simple
            # usage cases.
            # todo: Handle hollerith constants in DATA statement
            if _IS_CALL_STMT(line):
                l2 = self.line[4:].lstrip()
                i = l2.find('(')
                if i != -1 and l2[-1] == ')':
                    substrings = ['call '+l2[:i+1]]
                    start_search = _HOLLERITH_START_SEARCH
                    l2 = l2[i+1:-1].strip()
                    m = start_search(l2)
                    while m:
                        substrings.append(l2[:m.start()])
                        substrings.append(m.group('pre'))
                        num = int(m.group('num'))
                        substrings.append("'"+l2[m.end():m.end()+num]+"'")
                        l2 = l2[m.end()+num:]
                        m = start_search(l2)
                    substrings.append(l2)
                    substrings.append(')')
                    line = ''.join(substrings)

        line, str_map = string_replace_map(line,
                                           lower=not self.reader.format.is_pyf)
        self.strline = line
        self.strlinemap = str_map
        return line

    def parse_line(self, cls, parent_cls):
        if cls not in self.parse_cache:
            self.parse_cache[cls] = None
            obj = cls(self.line, parent_cls=parent_cls)
            self.parse_cache[cls] = obj
        else:
            obj = self.parse_cache[cls]
        return obj

    def parse_block(self, reader, cls, parent_cls):
        key = cls, tuple(parent_cls)
        if key not in self.parse_cache:
            obj = cls(reader, parent_cls=parent_cls)
            self.parse_cache[key] = obj
        else:
            obj = self.parse_cache[key]
        return obj


class SyntaxErrorLine(Line, FortranReaderError):
    '''
    Indicates a syntax error while processing a line.
    '''
    def __init__(self, line, linenospan, label, name, reader, message):
        Line.__init__(self, line, linenospan, label, name, reader)
        FortranReaderError.__init__(self, message)


class Comment(object):
    '''Holds a Fortran comment.

    :param str comment: String containing the text of a single or \
    multi-line comment
    :param linenospan: A 2-tuple containing the start and end line \
    numbers of the comment from the input source.
    :type linenospan: (int, int)
    :param reader: The reader object being used to read the input \
    source.
    :type reader: :py:class:`fparser.common.readfortran.FortranReaderBase`

    '''
    def __init__(self, comment, linenospan, reader):

        self.comment = comment
        self.span = linenospan
        self.reader = reader
        # self.line provides a common way to retrieve the content from
        # either a 'Line' or a 'Comment' class. This is useful for
        # tests as a reader can return an instance of either class and
        # we might want to check the contents in a consistent way.
        self.line = comment

    def __repr__(self):
        return self.__class__.__name__+'(%r,%s)' \
               % (self.comment, self.span)

    def isempty(self, ignore_comments=False):
        '''
        Whether or not this comment is in fact empty (or we are ignoring
        it). Provided for compatibility with Line.isempty()

        :param bool ignore_comments: whether we ignore comments
        :return: True if we are ignoring comments, False otherwise
        :rtype: bool
        '''
        return ignore_comments


class MultiLine(object):
    """ Holds PYF file multiline.

    PYF file multiline is represented as follows::
      prefix+'''+lines+'''+suffix.

    Attributes
    ----------
    prefix : str
    block : list
      list of lines
    suffix : str
    span : 2-tuple
      starting and ending line numbers
    reader : FortranReaderBase
    """
    def __init__(self, prefix, block, suffix, linenospan, reader):
        self.prefix = prefix
        self.block = block
        self.suffix = suffix
        self.span = linenospan
        self.reader = reader

    def __repr__(self):
        string = '{cls}({prefix!r},{block},{suffix!r},{span})'
        return string.format(cls=self.__class__.__name__,
                             prefix=self.prefix,
                             block=self.block,
                             suffix=self.suffix,
                             span=self.span)

    def isempty(self, ignore_comments=False):
        '''
        Returns true if there is no significant text in this multi-line
        string.
        '''
        return not (self.prefix or self.block or self.suffix)


class SyntaxErrorMultiLine(MultiLine, FortranReaderError):
    '''
    Indicates a syntax error while processing Python multi-line strings.
    '''
    def __init__(self, prefix, block, suffix, linenospan, reader, message):
        MultiLine.__init__(self, prefix, block, suffix, linenospan, reader)
        FortranReaderError.__init__(self, message)


class CppDirective(Line):
    '''Holds a preprocessor directive source line.

    :param str line: string containing the text of a single or \
                     multi-line preprocessor directive.
    :param linenospan: a 2-tuple containing the start and end line \
                       numbers of the directive from the input source.
    :type linenospan: (int, int)
    :param reader: The reader object being used to read the input \
                   source.
    :type reader: :py:class:`fparser.common.readfortran.FortranReaderBase`

    '''
    def __init__(self, line, linenospan, reader):
        super(CppDirective, self).__init__(
            line, linenospan, None, None, reader)

##############################################################################


class FortranReaderBase(object):
    """
    Base class for reading Fortran sources.

    A Fortran source must be a file-like object (have a ``.next()``
    method) and it may hold Fortran 77 code, fixed format Fortran
    code, free format Fortran code, or PYF signatures (with extended
    free format Fortran syntax).

    :param source: A file-like object with .next() method used to
                   retrive a line.
    :type source: either :py:class:`six.StringIO` or a file handle
    :param mode: A FortranFormat object as returned by
                 `sourceinfo.get_source_info()`
    :type mode: :py:class:`fparser.common.sourceinfo.Format`
    :param bool isstrict: Whether we are strictly enforcing fixed format
    :param bool ignore_comments: Whether or not to discard comments

    The Fortran source is iterated by `get_single_line`,
    `get_next_line`, `put_single_line` methods.

    """

    def __init__(self, source, mode, ignore_comments):
        self.source = source
        self._format = mode

        self.linecount = 0   # the current number of consumed lines
        self.isclosed = False
        # This value for ignore_comments can be overridden by using the
        # ignore_comments optional argument to e.g. get_single_line()
        self._ignore_comments = ignore_comments

        self.filo_line = []  # used for un-consuming lines.
        self.fifo_item = []
        self.source_lines = []  # source lines cache

        self.f2py_comment_lines = []  # line numbers of f2py directives

        self.reader = None
        self.include_dirs = ['.']

        self.source_only = None

        self.exit_on_error = True
        self.restore_cache = []

        return

    ##########################################################################

    def __repr__(self):
        return '%s(%r, %r, %r)' % (self.__class__.__name__,
                                   self.source,
                                   self._format.is_free,
                                   self._format.is_strict)

    def find_module_source_file(self, mod_name):
        '''
        Scans registered dependees for a named module.
        '''
        from .utils import get_module_file, module_in_file
        if self.source_only:
            for sf in self.source_only:
                if module_in_file(mod_name, sf):
                    return sf
        else:
            fn = None
            for d in self.include_dirs:
                fn = get_module_file(mod_name, d)
                if fn is not None:
                    return fn

    def set_format(self, mode):
        '''
        Set Fortran code mode (fixed/free format etc).

        :param mode: Object describing the desired mode for the reader
        :type mode: :py:class:`fparser.common.sourceinfo.FortranFormat`
        '''
        self._format = mode

    @property
    def format(self):
        '''
        :returns: the currently applicable format.
        :rtype: :py:class:`fparser.sourceinfo.FortranFormat`
        '''
        return self._format

    @property
    def name(self):
        '''
        :returns: the name of this reader.
        :rtype: str
        '''
        return '{source} mode={mode}'.format(source=self.source,
                                             mode=self._format.mode)

    def close_source(self):
        """ Called when self.source.next() raises StopIteration.
        """
        pass

    # For handling raw source lines:

    def put_single_line(self, line):
        """ Put single line to FILO line buffer.

        ``linecount`` will be decremented, that is, the line was
        returned by ``get_single_line`` call then it will be
        un-consumed.

        See also
        --------
        get_single_line, get_next_line
        """
        self.filo_line.append(line)
        self.linecount -= 1
        return

    def get_single_line(self, ignore_empty=False, ignore_comments=None):
        """ Return line from FILO line buffer or from source.

        First try getting the line from FILO line buffer.

        If FILO line buffer is empty then get the next line from
        source. Tabs in source line are expanded, ending blank and new
        line characters will be removed.  The source line will be
        added to ``source_lines`` list. If source line is empty then
        recursively get next non-empty line.

        In both situations ``linecount`` will be incremented, that is,
        the line will be consumed.

        :param bool ignore_empty: If True then ignore empty lines.
        :param bool ignore_comments: If True then ignore comments (overrides
                                     self._ignore_comments)

        See also
        --------
        put_single_line, get_next_line
        """
        if ignore_comments is None:
            ignore_comments = self._ignore_comments

        try:
            line = self.filo_line.pop()
            self.linecount += 1
            return line
        except IndexError:
            pass
        if self.isclosed:
            return None
        try:
            line = next(self.source)
        except StopIteration:
            self.isclosed = True
            self.close_source()
            return None
        self.linecount += 1

        if six.PY2 and not isinstance(line, six.text_type):
            # Ensure we always have a unicode object in Python 2.
            line = unicode(line, 'UTF-8')

        # expand tabs, replace special symbols, get rid of nl characters
        line = line.expandtabs().replace(u'\xa0', u' ').rstrip()

        if six.PY2:
            # Cast the unicode to str if we can do so safely. This
            # maximises compatibility with the existing Python 2 tests
            # and avoids the need to proliferate the use of unicode
            # literals (e.g. u"") in the parse tree repr.
            try:
                line = line.encode('ascii', errors='strict')
            except UnicodeEncodeError:
                # Can't cast to str as there are non-ascii characters
                # in the line.
                pass

        self.source_lines.append(line)

        if ignore_comments and (self.format.is_fixed or self.format.is_f77):
            # Check for a fixed-format comment. If the current line *is*
            # a comment and we are ignoring them, then recursively call this
            # routine again to get the next source line.
            if _is_fix_comment(line, isstrict=self.format.is_strict):
                return self.get_single_line(ignore_empty, ignore_comments)

        if ignore_empty and not line:
            return self.get_single_line(ignore_empty, ignore_comments)

        return line

    def get_next_line(self, ignore_empty=False, ignore_comments=None):
        """ Return next non-empty line from FILO line buffer or from source.

        The line will be put to FILO line buffer. So, this method can
        be used for looking forward lines without consuming them.

        See also
        --------
        get_single_line, put_single_line
        """
        if ignore_comments is None:
            ignore_comments = self._ignore_comments

        line = self.get_single_line(ignore_empty, ignore_comments)
        if line is None:
            return
        self.put_single_line(line)
        return line

    # Parser methods:
    def get_item(self, ignore_comments=None):
        """ Return next item.
        """
        if ignore_comments is None:
            ignore_comments = self._ignore_comments

        try:
            item = self.next(ignore_comments=ignore_comments)
        except StopIteration:
            return
        return item

    def put_item(self, item):
        """ Insert item to FIFO item buffer.
        """
        self.fifo_item.insert(0, item)
        return

    # Iterator methods:

    def __iter__(self):
        """ Make FortranReader an iterator.
        """
        return self

    def __next__(self):
        return self.next()

    def next(self, ignore_comments=None):
        '''Return the next Fortran code item. Include statements are dealt
        with here.

        :param bool ignore_comments: When True then act as if Fortran \
        code does not contain any comments or blank lines. if this \
        optional arguement is not provided then use the default \
        value.

        :returns: the next line item. This can be from a local fifo \
        buffer, from an include reader or from this reader.
        :rtype: py:class:`fparser.common.readfortran.Line`

        :raises StopIteration: if no more lines are found.
        :raises StopIteration: if a general error has occured.

        '''
        if ignore_comments is None:
            ignore_comments = self._ignore_comments
        try:
            if self.reader is not None:
                # inside INCLUDE statement
                try:
                    # Manually check to see if something has not
                    # matched and has been placed in the fifo. We
                    # can't use _next() as this method is associated
                    # with the include reader (self.reader._next()),
                    # not this reader (self._next()).
                    return self.fifo_item.pop(0)
                except IndexError:
                    # There is nothing in the fifo buffer.
                    try:
                        # Return a line from the include.
                        return self.reader.next(ignore_comments)
                    except StopIteration:
                        # There is nothing left in the include
                        # file. Setting reader to None indicates that
                        # we should now read from the main reader.
                        self.reader = None
            item = self._next(ignore_comments)
            if isinstance(item, Line) and _IS_INCLUDE_LINE(item.line):
                # catch INCLUDE statement and create a new FortranReader
                # to enter to included file.
                reader = item.reader
                filename = item.line.strip()[7:].lstrip()[1:-1]
                include_dirs = self.include_dirs[:]
                path = filename
                for incl_dir in include_dirs:
                    path = os.path.join(incl_dir, filename)
                    if os.path.exists(path):
                        break
                if not os.path.isfile(path):
                    # The include file does not exist in the specified
                    # locations.
                    #
                    # The Fortran standard states that an INCLUDE line
                    # is not a Fortran statement. However, fparser is
                    # a parser not a compiler and some subsequent tool
                    # might need to make use of this include so we
                    # return it and let the parser deal with it.
                    #
                    return item
                reader.info('including file %r' % (path), item)
                self.reader = FortranFileReader(
                    path,
                    include_dirs=include_dirs,
                    ignore_comments=ignore_comments)
                result = self.reader.next(ignore_comments=ignore_comments)
                return result
            return item
        except StopIteration:
            raise
        # TODO can we specify one or more specific exception types
        # rather than catching *every* exception.
        except Exception as err:
            message = self.format_message('FATAL ERROR',
                                          'while processing line',
                                          self.linecount, self.linecount)
            logging.getLogger(__name__).critical(message)
            message = 'Traceback\n' + ''.join(traceback.format_stack())
            logging.getLogger(__name__).debug(message)
            logging.getLogger(__name__).debug(six.text_type(err))
            logging.getLogger(__name__).critical('STOPPED READING')
            raise StopIteration

    def _next(self, ignore_comments=None):
        """
        Return the next item from FIFO item buffer or construct
        one from source line.

        Resolves ``;`` statement terminations.

        :param bool ignore_comments: Whether or not to ignore comments
                                     (overrides self._ignore_comments)

        See also
        --------
        next, get_source_item
        """
        if ignore_comments is None:
            ignore_comments = self._ignore_comments
        fifo_item_pop = self.fifo_item.pop
        while 1:
            try:
                # first empty the FIFO item buffer:
                item = fifo_item_pop(0)
            except IndexError:
                # construct a new item from source
                item = self.get_source_item()
                if item is None:
                    raise StopIteration
            if not (item.isempty(ignore_comments)):
                break
            # else ignore empty lines and comments by getting next line

        if not isinstance(item, Comment):
            # resolve `;` statement terminations
            if not self._format.is_pyf and isinstance(item, Line) \
                   and not item.is_f2py_directive \
                   and ';' in item.get_line():
                # ;-separator not recognized in pyf-mode
                items = []
                # Deal with each Fortran statement separately.
                split_line_iter = iter(item.get_line().split(';'))
                first = next(split_line_iter)
                # The full line has already been processed as a Line
                # object in 'item' (and may therefore have label
                # and/or construct name properties extracted from the
                # start of the line). The simplest way to avoid losing
                # any label or construct name properties for the first
                # statement is to copy the 'item' object and update it
                # so that it only includes text for the first
                # statement (rather than the full line). Subsequent
                # statements need to be processed into Line
                # objects.
                items.append(item.copy(first.strip(), apply_map=True))
                for line in split_line_iter:
                    # Any subsequent statements have not been processed
                    # before, so new Line objects need to be created.
                    line = line.strip()
                    if line:
                        # The statement might have a label and/or construct
                        # name.
                        label, line = extract_label(line)
                        name, line = extract_construct_name(line)
                        # Create a new Line object and append to items
                        # using the existing span (line numbers) and
                        # reader.
                        new_line = Line(item.apply_map(line), item.span, label,
                                        name, item.reader)
                        items.append(new_line)
                items.reverse()
                for newitem in items:
                    self.fifo_item.insert(0, newitem)
                return fifo_item_pop(0)
        return item

    # Interface to returned items:

    def line_item(self,
                  line,
                  startlineno,
                  endlineno,
                  label,
                  name,
                  errmessage=None):
        """ Construct Line item.
        """
        if errmessage is None:
            return Line(line, (startlineno, endlineno), label, name, self)
        return SyntaxErrorLine(line, (startlineno, endlineno),
                               label, name, self, errmessage)

    def multiline_item(self, prefix, lines, suffix,
                       startlineno, endlineno, errmessage=None):
        """ Construct MultiLine item.
        """
        if errmessage is None:
            return MultiLine(prefix,
                             lines,
                             suffix,
                             (startlineno, endlineno),
                             self)
        return SyntaxErrorMultiLine(prefix,
                                    lines,
                                    suffix,
                                    (startlineno, endlineno),
                                    self,
                                    errmessage)

    def comment_item(self, comment, startlineno, endlineno):
        """ Construct Comment item.
        """
        return Comment(comment, (startlineno, endlineno), self)

    def cpp_directive_item(self, line, startlineno, endlineno):
        '''
        Construct :py:class:`fparser.common.readfortran.CppDirective` item.

        :param str line: string containing the text of a single or \
                         multi-line preprocessor directive.
        :param int startlineno: start line number of the directive from \
                                the input source.
        :param int endlineno: end line number of the directive from \
                              the input source.

        '''
        return CppDirective(line, (startlineno, endlineno), self)

    # For handling messages:

    def format_message(self, kind, message, startlineno, endlineno,
                       startcolno=0, endcolno=-1):
        '''
        Prepares a string for logging.
        '''
        back_index = {'warning': 2,
                      'error': 3,
                      'info': 0}.get(kind.lower(), 3)
        r = ['While processing %r (mode=%r)..' % (self.id,
                                                  self._format.mode)]
        for i in range(max(1, startlineno - back_index), startlineno):
            r.append('%5d:%s' % (i, self.source_lines[i - 1]))
        for i in range(startlineno, min(endlineno+back_index,
                                        len(self.source_lines))+1):
            if i == 0 and not self.source_lines:
                break
            linenostr = '%5d:' % (i)
            if i == endlineno:
                sourceline = self.source_lines[i-1]
                l0 = linenostr+sourceline[:startcolno]
                if endcolno == -1:
                    l1 = sourceline[startcolno:]
                    l2 = ''
                else:
                    l1 = sourceline[startcolno:endcolno]
                    l2 = sourceline[endcolno:]
                r.append('%s%s%s <== %s' % (l0, l1, l2, message))
            else:
                r.append(linenostr + self.source_lines[i-1])
        return '\n'.join(r)

    def format_error_message(self, message, startlineno, endlineno,
                             startcolno=0, endcolno=-1):
        '''Create a string with an error message.'''
        return self.format_message('ERROR', message, startlineno,
                                   endlineno, startcolno, endcolno)

    def format_warning_message(self, message, startlineno, endlineno,
                               startcolno=0, endcolno=-1):
        '''Create a string with a warning message. '''
        return self.format_message('WARNING', message, startlineno,
                                   endlineno, startcolno, endcolno)

    def info(self, message, item=None):
        '''
        Logs an information message.
        '''
        if item is None:
            m = self.format_message('INFORMATION',
                                    message,
                                    len(self.source_lines)-2,
                                    len(self.source_lines))
        else:
            m = self.format_message('INFORMATION',
                                    message,
                                    item.span[0], item.span[1])
        logging.getLogger(__name__).info(m)
        return

    def error(self, message, item=None):
        '''
        Logs an error message.
        '''
        if item is None:
            m = self.format_error_message(message, len(self.source_lines)-2,
                                          len(self.source_lines))
        else:
            m = self.format_error_message(message, item.span[0], item.span[1])
        logging.getLogger(__name__).error(m)
        if self.exit_on_error:
            sys.exit(1)
        return

    def warning(self, message, item=None):
        '''
        Logs a warning message.
        '''
        if item is None:
            m = self.format_warning_message(message,
                                            len(self.source_lines) - 2,
                                            len(self.source_lines))
        else:
            m = self.format_warning_message(message,
                                            item.span[0],
                                            item.span[1])
        logging.getLogger(__name__).warning(m)
        return

    # Auxiliary methods for processing raw source lines:

    def handle_cpp_directive(self, line):
        '''
        Determine whether the current line is likely to hold
        C preprocessor directive.

        If the first non whitespace character of a line is the symbol ``#``
        it is assumed this is a preprocessor directive.

        Preprocessor directives can be used only in Fortran codes. They are
        ignored when used inside PYF files.

        The actual line content is not altered.

        :param str line: the line to be tested for directives.

        :return: a tuple containing the line and True/False depending on \
                 whether it holds a C preprocessor directive.
        :rtype: (str, bool)

        '''
        if not line or self._format.is_pyf:
            return (line, False)
        return (line, line.lstrip().startswith('#'))

    def handle_cf2py_start(self, line):
        """ Apply f2py directives to line.

        F2py directives are specified in the beginning of the line.

        f2py directives can be used only in Fortran codes.  They are
        ignored when used inside PYF files.

        Parameters
        ----------
        line : str

        Returns
        -------
        line : str
        """
        if not line or self._format.is_pyf:
            return line
        if self._format.is_fixed:
            if line[0] in '*cC!#':
                if line[1:5].lower() == 'f2py':
                    line = 5*' ' + line[5:]
                    self.f2py_comment_lines.append(self.linecount)
            if self._format.is_f77:
                return line
        m = _CF2PY_RE.match(line)
        if m:
            newline = m.group('indent')+5*' '+m.group('rest')
            self.f2py_comment_lines.append(self.linecount)
            assert len(newline) == len(line), repr((newline, line))
            return newline
        return line

    def handle_inline_comment(self, line, lineno, quotechar=None):
        """
        Any in-line comment is extracted from the line. If
        keep_inline_comments==True then the extracted comments are put back
        into the fifo sequence where they will subsequently be processed as
        a comment line.

        :param str line: line of code from which to remove in-line comment
        :param int lineno: line-no. in orig. file
        :param quotechar: String to use as character-string delimiter
        :type quotechar: {None, str}

        :return: line_with_no_comments, quotechar, had_comment
        :rtype: 3-tuple of str, str, bool
        """
        had_comment = False
        if quotechar is None and '!' not in line and \
           '"' not in line and "'" not in line:
            # There's no comment on this line
            return line, quotechar, had_comment

        idx = line.find('!')
        put_item = self.fifo_item.append
        if quotechar is None and idx != -1:
            # first try a quick method:
            newline = line[:idx]
            if '"' not in newline and '\'' not in newline:
                if self.format.is_f77 or not line[idx:].startswith('!f2py'):
                    put_item(self.comment_item(line[idx:], lineno, lineno))
                    return newline, quotechar, True

        # We must allow for quotes...
        items, newquotechar = splitquote(line, quotechar)
        noncomment_items = []
        noncomment_items_append = noncomment_items.append
        n = len(items)
        commentline = None
        for k in range(n):
            item = items[k]
            if isinstance(item, String) or '!' not in item:
                noncomment_items_append(item)
                continue
            j = item.find('!')
            noncomment_items_append(item[:j])
            items[k] = item[j:]
            commentline = ''.join(items[k:])
            break
        if commentline is not None:
            if commentline.startswith('!f2py'):
                # go to next iteration:
                newline = ''.join(noncomment_items) + commentline[5:]
                self.f2py_comment_lines.append(lineno)
                return self.handle_inline_comment(newline, lineno, quotechar)
            put_item(self.comment_item(commentline, lineno, lineno))
            had_comment = True
        return ''.join(noncomment_items), newquotechar, had_comment

    def handle_multilines(self, line, startlineno, mlstr):
        '''
        Examines line for Python triple quote strings (f2py feature).

        :param str line: line of Fortran source text
        :param int startlineno: the number of the line on which this
                                 multi-line string began.
        :param list mlstr: list of delimiters for a multi-line string
                           (e.g. '"""')
        '''
        i = line.find(mlstr)
        if i != -1:
            prefix = line[:i]
            # skip fake multiline starts
            p, k = prefix, 0
            while p.endswith('\\'):
                p, k = p[:-1], k + 1
            if k % 2:
                return
        if i != -1 and '!' not in prefix:
            # Note character constants like 'abc"""123',
            # so multiline prefix should better not contain `'' or `"' not `!'.
            for quote in '"\'':
                if prefix.count(quote) % 2:
                    message = 'multiline prefix contains odd number of' \
                              + ' {!r} characters'.format(quote)
                    message = self.format_warning_message(message,
                                                          startlineno,
                                                          startlineno,
                                                          0,
                                                          len(prefix))
                    logging.getLogger(__name__).warning(message)

            suffix = None
            multilines = []
            line = line[i+3:]
            while line is not None:
                j = line.find(mlstr)
                if j != -1 and '!' not in line[:j]:
                    multilines.append(line[:j])
                    suffix = line[j+3:]
                    break
                multilines.append(line)
                line = self.get_single_line()
            if line is None:
                message = 'multiline block never ends'
                message = self.format_error_message(message,
                                                    startlineno,
                                                    startlineno,
                                                    i)
                return self.multiline_item(prefix,
                                           multilines,
                                           suffix,
                                           startlineno,
                                           self.linecount,
                                           message)
            suffix, qc, had_comment \
                = self.handle_inline_comment(suffix, self.linecount)
            # no line continuation allowed in multiline suffix
            if qc is not None:
                message = 'following character continuation: {!r},' \
                          + ' expected None.'
                message = self.format_message('ASSERTION FAILURE(pyf)',
                                              message.format(qc),
                                              startlineno,
                                              self.linecount)
                logging.getLogger(__name__).warning(message)
            # XXX: should we do line.replace('\\'+mlstr[0],mlstr[0])
            #      for line in multilines?
            return self.multiline_item(prefix,
                                       multilines,
                                       suffix,
                                       startlineno,
                                       self.linecount)

    # The main method of interpreting raw source lines within
    # the following contexts: f77, fixed, free, pyf.

    def get_source_item(self):
        """ Return next source item.

        A source item is ..
        - a fortran line
        - a list of continued fortran lines
        - a multiline - lines inside triple-quotes, only when in ispyf mode
        - a comment line
        - a preprocessor directive line
        """
        get_single_line = self.get_single_line
        line = get_single_line()
        if line is None:
            return
        startlineno = self.linecount
        line, is_cpp_directive = self.handle_cpp_directive(line)
        if is_cpp_directive:
            # CPP directive line
            lines = []
            while line.rstrip().endswith('\\'):
                # Line continuation
                lines.append(line.rstrip()[:-1])
                line = get_single_line()
            lines.append(line)
            endlineno = self.linecount
            return self.cpp_directive_item(''.join(lines), startlineno,
                                           endlineno)

        line = self.handle_cf2py_start(line)
        is_f2py_directive = startlineno in self.f2py_comment_lines
        isstrict = self._format.is_strict
        have_comment = False
        label = None
        name = None

        if self._format.is_pyf:
            # handle multilines
            for mlstr in ['"""', "'''"]:
                r = self.handle_multilines(line, startlineno, mlstr)
                if r:
                    return r
        if self._format.is_fixed:
            if _is_fix_comment(line, isstrict):
                # comment line:
                return self.comment_item(line, startlineno, startlineno)

            for i in range(min(5, len(line))):
                # check that fixed format line starts according to Fortran
                # standard
                if line[i] not in _SPACEDIGITS:
                    message = 'non-space/digit char %r found in column %i'\
                              ' of fixed Fortran code' % (line[i], i + 1)
                    if i == 0:
                        message += ', interpreting line as comment line'
                    if self._format.is_fix:
                        if i != 0:
                            message += ', switching to free format mode'
                        message = self.format_warning_message(message,
                                                              startlineno,
                                                              self.linecount)
                        logging.getLogger(__name__).warning(message)
                        if i == 0:
                            # non standard comment line:
                            return self.comment_item(line,
                                                     startlineno,
                                                     startlineno)
                        mode = fparser.common.sourceinfo.FortranFormat(True,
                                                                       False)
                        self.set_format(mode)
                    else:
                        message = self.format_warning_message(message,
                                                              startlineno,
                                                              self.linecount)
                        logging.getLogger(__name__).warning(message)
                        if i == 0:
                            # non standard comment line:
                            return self.comment_item(line,
                                                     startlineno,
                                                     startlineno)
                        # return line item with error message
                        # TODO: handle cases with line[6:]==''
                        message = self.format_error_message(message,
                                                            startlineno,
                                                            self.linecount)
                        return self.line_item(line[6:],
                                              startlineno,
                                              self.linecount,
                                              label,
                                              name,
                                              message)
            if self._format.is_fixed:  # Check for switched to free format
                # check for label
                s = line[:5].strip().lower()
                if s:
                    label = int(s)
                if not self._format.is_f77:
                    m = _CONSTRUCT_NAME_RE.match(line[6:])
                    if m:
                        name = m.group('name')
                        line = line[:6] + line[6:][m.end():].lstrip()
                if not line[6:].strip():
                    # check for a blank line
                    if name is not None:
                        self.error('No construct following construct-name.')
                    elif label is not None:
                        self.warning('Label must follow nonblank character'
                                     + ' (F2008:3.2.5_2)')
                    return self.comment_item('', startlineno, self.linecount)
                # line is not a comment and the start of the line is valid

        if self._format.is_f77 and not is_f2py_directive:
            # Fortran 77 is easy..
            lines = [line[6:72]]
            # get_next_line does not actually consume lines - they are put
            # into the FILO buffer as well as being returned. This means
            # that we can ignore comments for the purposes of dealing
            # with the continued line and then handle them as though they
            # follow on after the single line constructed from the multiple
            # continued lines.
            while _is_fix_cont(self.get_next_line(ignore_empty=True,
                                                  ignore_comments=True)):
                # handle fix format line continuations for F77 code
                line = get_single_line()
                lines.append(line[6:72])
            return self.line_item(''.join(lines), startlineno,
                                  self.linecount, label, name)

        handle_inline_comment = self.handle_inline_comment

        endlineno = self.linecount
        if self._format.is_fix and not is_f2py_directive:
            # handle inline comment
            newline, qc, had_comment = handle_inline_comment(line[6:],
                                                             startlineno)
            have_comment |= had_comment
            lines = [newline]
            next_line = self.get_next_line()

            while _is_fix_cont(next_line) or _is_fix_comment(next_line,
                                                             isstrict):
                # handle fix format line continuations for F90 or
                # newer code.  Mixing fix format and free format line
                # continuations is not allowed nor detected, just
                # eject warnings.
                line2 = get_single_line()  # consume next_line as line2
                if _is_fix_comment(line2, isstrict):
                    # handle fix format comments inside line continuations
                    # after the line construction
                    citem = self.comment_item(line2,
                                              self.linecount,
                                              self.linecount)
                    self.fifo_item.append(citem)
                else:
                    # line continuation
                    newline, qc, had_comment \
                        = self.handle_inline_comment(line2[6:],
                                                     self.linecount,
                                                     qc)
                    have_comment |= had_comment
                    lines.append(newline)
                    endlineno = self.linecount
                next_line = self.get_next_line()
            # no character continuation should follows now
            if qc is not None:
                message = 'following character continuation: ' \
                          + '{!r}, expected None.'
                message = self.format_message('ASSERTION FAILURE(fix)',
                                              message.format(qc),
                                              startlineno,
                                              self.linecount)
                logging.getLogger(__name__).warning(message)
            if len(lines) > 1:
                for i in range(len(lines)):
                    line = lines[i]
                    if line.rstrip().endswith('&'):
                        message = 'free format line continuation character ' \
                                  + "`&' detected in fix format code"
                        location = line.rfind('&') + 5
                        message = self.format_warning_message(message,
                                                              startlineno + i,
                                                              startlineno + i,
                                                              location)
                        logging.getLogger(__name__).warning(message)
            return self.line_item(''.join(lines),
                                  startlineno,
                                  endlineno,
                                  label,
                                  name)

        # line is free format or fixed format with f2py directive (that
        # will be interpretted as free format line).

        start_index = 0
        if self._format.is_fix:
            start_index = 6
        lines = []
        lines_append = lines.append
        put_item = self.fifo_item.append
        qc = None
        while line is not None:
            if start_index:  # fix format code
                line, qc, had_comment \
                    = handle_inline_comment(line[start_index:],
                                            self.linecount, qc)
                have_comment |= had_comment
                is_f2py_directive = self.linecount in self.f2py_comment_lines
            else:
                # free format
                line_lstrip = line.lstrip()
                if lines:
                    if line_lstrip.startswith('!'):
                        # check for comment line within line continuation
                        put_item(self.comment_item(line_lstrip,
                                                   self.linecount,
                                                   self.linecount))
                        have_comment = True
                        line = get_single_line()
                        continue
                    elif line_lstrip == "":
                        # skip blank lines within a line continuation
                        line = get_single_line()
                        continue
                else:
                    # Extract label and/or construct name from line if
                    # there is one.
                    label, line = extract_label(line)
                    name, line = extract_construct_name(line)

                line, qc, had_comment = \
                    handle_inline_comment(line, self.linecount, qc)
                have_comment |= had_comment
                is_f2py_directive = self.linecount in self.f2py_comment_lines

            i = line.rfind('&')
            if i != -1:
                line_i1_rstrip = line[i+1:].rstrip()
            if not lines:
                # first line
                if i == -1 or line_i1_rstrip:
                    lines_append(line)
                    break
                endlineno = self.linecount
                lines_append(line[:i])
                line = get_single_line()
                continue
            if i == -1 or line_i1_rstrip:
                # no line continuation follows
                i = len(line)
            k = -1
            if i != -1:
                # handle the beggining of continued line
                k = line[:i].find('&')
                if k != 1 and line[:k].lstrip():
                    k = -1
            endlineno = self.linecount
            lines_append(line[k+1:i])
            if i == len(line):
                break
            line = get_single_line()

        if qc is not None:
            message = 'following character continuation: {!r}, ' \
                      + 'expected None.'
            message = self.format_message('ASSERTION FAILURE(free)',
                                          message.format(qc),
                                          startlineno,
                                          endlineno)
            logging.getLogger(__name__).error(message)
        line_content = ''.join(lines).strip()
        if line_content:
            return self.line_item(line_content,
                                  startlineno,
                                  endlineno,
                                  label,
                                  name)
        if label is not None:
            message = 'Label must follow nonblank character (F2008:3.2.5_2)'
            self.warning(message)
        if name is not None:
            self.error('No construct following construct-name.')
        if have_comment:
            return next(self)
        return self.comment_item('', startlineno, endlineno)


class FortranFileReader(FortranReaderBase):
    '''
    Constructs a FortranFileReader object from a file.

    :param file_candidate: A filename or file-like object.
    :param list include_dirs: Directories in which to look for inclusions.
    :param list source_only: Fortran source files to search for modules
                             required by "use" statements.
    :param bool ignore_comments: Whether or not to ignore comments

    For example:

    >>> from fparser.common.readfortran import FortranFileReader
    >>> import os
    >>> reader = FortranFileReader(\'myfile.f90\')

    '''
    def __init__(self, file_candidate, include_dirs=None, source_only=None,
                 ignore_comments=True):
        # The filename is used as a unique ID. This is then used to cache the
        # contents of the file. Obviously if the file changes content but not
        # filename, problems will ensue.
        #
        self._remove_on_destruction = False
        self._close_on_destruction = False
        if isinstance(file_candidate, six.string_types):
            self.id = file_candidate
            from fparser.common.utils import make_clean_tmpfile
            # Handle potential invalid characters in the input. Done
            # by creating a new file (tmpfile) with any errors removed
            # (or raising an exception - see make_clean_tmpfile).
            tmpfile = make_clean_tmpfile(file_candidate)
            self.file = io.open(tmpfile, 'r', encoding='UTF-8')
            self._close_on_destruction = True
            self._remove_on_destruction = True
        elif hasattr(file_candidate,
                     'read') and hasattr(file_candidate,
                                         'name'):  # Is likely a file
            self.id = file_candidate.name
            self.file = file_candidate
        else:  # Probably not something we can deal with
            message = 'FortranFileReader is used with a filename'
            message += ' or file-like object.'
            raise ValueError(message)
        mode = fparser.common.sourceinfo.get_source_info(file_candidate)

        FortranReaderBase.__init__(self, self.file, mode, ignore_comments)

        if include_dirs is None:
            self.include_dirs.insert(0, os.path.dirname(self.id))
        else:
            self.include_dirs = include_dirs[:]
        if source_only is not None:
            self.source_only = source_only[:]
        return

    def __del__(self):
        if self._close_on_destruction:
            self.file.close()
            if self._remove_on_destruction:
                os.remove(self.file.name)

    def close_source(self):
        self.file.close()


class FortranStringReader(FortranReaderBase):
    '''
    Reads Fortran source code as a string.

    :param str string: string to read
    :param list include_dirs: List of dirs to search for include files
    :param list source_only: Fortran source files to search for modules
                             required by "use" statements.
    :param bool ignore_comments: Whether or not to ignore comments

    For example:

    >>> from fparser.common.readfortran import FortranStringReader
    >>> code = \'\'\'
             subroutine foo(a)
                integer a
                print*,\"a=\",a
              end
        \'\'\'
    >>> reader = FortranStringReader(code)

    '''
    def __init__(self, string, include_dirs=None, source_only=None,
                 ignore_comments=True):
        # The Python ID of the string was used to uniquely identify it for
        # caching purposes. Unfortunately this ID is only unique for the
        # lifetime of the string. In CPython it is the address of the string
        # and the chance of a new string being allocated to the same address
        # is actually quite high. Particularly in a unit-testing scenario.
        #
        # For this reason the hash is used instead. A much better solution
        # anyway.
        #
        self.id = 'string-' + str(hash(string))
        source = six.StringIO(string)
        mode = fparser.common.sourceinfo.get_source_info_str(string)
        FortranReaderBase.__init__(self, source, mode,
                                   ignore_comments)
        if include_dirs is not None:
            self.include_dirs = include_dirs[:]
        if source_only is not None:
            self.source_only = source_only[:]
        return
